from random import shuffle
import unittest

import kopf

from koreo import result


class TestCombineMessages(unittest.TestCase):
    def test_no_outcomes(self):
        outcomes: list[result.Outcome] = []

        combined = result.combine(outcomes)

        self.assertIsInstance(combined, result.Skip)

    def test_depskips(self):
        outcomes: list[result.Outcome] = [
            result.DepSkip(),
            result.DepSkip(),
            result.DepSkip(),
            result.DepSkip(),
            result.DepSkip(),
        ]
        shuffle(outcomes)

        combined = result.combine(outcomes)

        self.assertIsInstance(combined, result.DepSkip)

    def test_skips(self):
        outcomes: list[result.Outcome] = [
            result.Skip(),
            result.Skip(),
            result.Skip(),
            result.Skip(),
            result.Skip(),
        ]
        shuffle(outcomes)

        combined = result.combine(outcomes)

        self.assertIsInstance(combined, result.Skip)

    def test_depskips_and_skips(self):
        outcomes: list[result.Outcome] = [
            result.DepSkip(),
            result.Skip(),
            result.DepSkip(),
            result.Skip(),
        ]
        shuffle(outcomes)

        combined = result.combine(outcomes)

        self.assertIsInstance(combined, result.Skip)

    def test_oks(self):
        outcomes: list[result.Outcome] = [
            result.Ok("test"),
            result.Ok(8),
            result.Ok(88),
            result.Ok(None),
            result.Ok(True),
        ]
        shuffle(outcomes)

        combined = result.combine(outcomes)

        self.assertIsInstance(combined, result.Ok)

        # Note, this is done so that we can shuffle the outcomes so that we're
        # testing different combinations of outcome orderings.
        for value in [None, "test", 8, 88, True]:
            self.assertIn(value, combined.data)

    def test_skips_and_oks(self):
        outcomes: list[result.Outcome] = [
            result.DepSkip(),
            result.DepSkip(),
            result.Ok("test"),
            result.Ok(8),
            result.Ok(False),
            result.Ok(None),
            result.Skip(),
            result.Skip(),
        ]
        shuffle(outcomes)

        combined = result.combine(outcomes)

        self.assertIsInstance(combined, result.Ok)

        # Note, this is done so that we can shuffle the outcomes so that we're
        # testing different combinations of outcome orderings.
        for value in [None, "test", 8, False]:
            self.assertIn(value, combined.data)

    def test_combine_one_value(self):
        outcomes: list[result.Outcome] = [
            result.Ok("ok-value"),
        ]

        combined = result.combine(outcomes)

        self.assertIsInstance(combined, result.Ok)

        print(combined)
        self.assertListEqual(["ok-value"], combined.data)

    def test_unwrapped_combine_one_value(self):
        outcomes: list[result.UnwrappedOutcome] = [
            "one-ok-value",
        ]

        combined = result.unwrapped_combine(outcomes)

        print(combined)
        self.assertListEqual(["one-ok-value"], combined)

    def test_retries(self):
        delay_5_message = "Waiting"
        default_delay_message = "Will retry"

        outcomes: list[result.Outcome] = [
            result.Retry(delay=500),
            result.Retry(delay=5, message="Waiting"),
            result.Retry(delay=59),
            result.Retry(message="Will retry"),
        ]
        shuffle(outcomes)

        combined = result.combine(outcomes)

        self.assertIsInstance(combined, result.Retry)
        self.assertEqual(500, combined.delay)
        self.assertIn(delay_5_message, combined.message)
        self.assertIn(default_delay_message, combined.message)

    def test_skips_oks_and_retries(self):
        outcomes: list[result.Outcome] = [
            result.DepSkip(),
            result.Skip(),
            result.Ok(None),
            result.Ok("test"),
            result.Ok(8),
            result.Retry(delay=500),
            result.Retry(delay=5, message="Waiting"),
        ]
        shuffle(outcomes)

        combined = result.combine(outcomes)

        self.assertIsInstance(combined, result.Retry)
        self.assertEqual(500, combined.delay)
        self.assertEqual("Waiting", combined.message)

    def test_permfail(self):
        first_message = "A bad error"
        second_message = "A really, really bad error"

        outcomes: list[result.Outcome] = [
            result.PermFail(),
            result.PermFail(message=first_message),
            result.PermFail(message=second_message),
            result.PermFail(),
        ]
        shuffle(outcomes)

        combined = result.combine(outcomes)

        self.assertIsInstance(combined, result.PermFail)
        self.assertIn(first_message, combined.message)
        self.assertIn(second_message, combined.message)

    def test_skips_oks_retries_and_permfail(self):
        outcomes: list[result.Outcome] = [
            result.DepSkip(),
            result.Skip(),
            result.Ok(None),
            result.Ok("test"),
            result.Ok(8),
            result.Retry(delay=500),
            result.Retry(delay=5, message="Waiting"),
            result.PermFail(),
            result.PermFail(message="All done"),
        ]
        shuffle(outcomes)

        combined = result.combine(outcomes)

        self.assertIsInstance(combined, result.PermFail)
        self.assertEqual("All done", combined.message)


class TestIsOk(unittest.TestCase):
    def test_skip(self):
        self.assertFalse(result.is_ok(result.Skip()))

    def test_ok(self):
        self.assertTrue(result.is_ok(result.Ok(None)))

    def test_retry(self):
        self.assertFalse(result.is_ok(result.Retry()))

    def test_permfail(self):
        self.assertFalse(result.is_ok(result.PermFail()))


class TestIsNotError(unittest.TestCase):
    def test_skip(self):
        self.assertTrue(result.is_not_error(result.Skip()))

    def test_ok(self):
        self.assertTrue(result.is_not_error(result.Ok(None)))

    def test_retry(self):
        self.assertFalse(result.is_not_error(result.Retry()))

    def test_permfail(self):
        self.assertFalse(result.is_not_error(result.PermFail()))


class TestIsError(unittest.TestCase):
    def test_skip(self):
        self.assertFalse(result.is_error(result.Skip()))

    def test_ok(self):
        self.assertFalse(result.is_error(result.Ok(None)))

    def test_retry(self):
        self.assertTrue(result.is_error(result.Retry()))

    def test_permfail(self):
        self.assertTrue(result.is_error(result.PermFail()))


class TestRaiseForError(unittest.TestCase):
    def test_retry(self):
        with self.assertRaisesRegex(kopf.TemporaryError, "Try again"):
            result.raise_for_error(result.Retry(message="Try again"))

    def test_permfail(self):
        with self.assertRaisesRegex(kopf.PermanentError, "Failure"):
            result.raise_for_error(result.PermFail(message="Failure"))



from lark import Tree, Token


def extract_argument_structure(compiled: Tree) -> set[str]:
    var_map: set[str] = set()
    for thing in compiled.iter_subtrees():
        if thing.data == "member_dot":
            member_dot: str = _process_member_dot(thing)
            var_map.add(member_dot)

        if thing.data == "member_index":
            member_index: str = _process_member_index(thing)
            var_map.add(member_index)

    return var_map


def _process_member_dot(tree: Tree):
    if len(tree.children) != 2:
        # TODO: Not sure this is possible?
        raise Exception(f"UNKNOWN MEMBER_DOT LENGTH! {len(tree.children)}: {tree}")

    # print(f"{len(tree.children)}: {tree.children[0]}")

    terminal = f"{tree.children[1]}"

    root: Tree = tree.children[0].children[0]

    if root.data == "member_dot":
        return f"{_process_member_dot(root)}.{terminal}"

    if root.data == "member_index":
        return f"{_process_member_index(root)}.{terminal}"

    if root.data == "member_dot_arg":
        return f"{_process_member_dot_arg(root)}.{terminal}"

    if root.data == "primary":
        return f"{_process_primary(root)}.{terminal}"

    # TODO: Is this possible?
    raise Exception(f"UNKNOWN MEMBER_DOT ROOT TYPE! {root}")


def _process_member_dot_arg(tree: Tree):
    if len(tree.children) != 3:
        # TODO: Not sure this is possible?
        raise Exception(f"UNKNOWN MEMBER_DOT_ARG LENGTH! {len(tree.children)}: {tree}")

    # print(f"{len(tree.children)}: {tree.children[0]}")

    terminal = f"{tree.children[1]}"

    root: Tree = tree.children[0].children[0]

    if root.data == "member_dot":
        return f"{_process_member_dot(root)}.{terminal}"

    if root.data == "member_index":
        return f"{_process_member_index(root)}.{terminal}"

    if root.data == "member_dot_arg":
        return f"{_process_member_dot_arg(root)}.{terminal}"

    if root.data == "primary" or root.data == "ident":
        return f"{_process_primary(root)}.{terminal}"

    # TODO: Is this possible?
    raise Exception(f"UNKNOWN MEMBER_DOT_ARG ROOT TYPE! {root}")


def _process_member_index(tree: Tree):
    if len(tree.children) != 2:
        # TODO: Not sure this is possible?
        raise Exception(f"UNKNOWN MEMBER_INDEX LENGTH! {len(tree.children)}: {tree}")

    # print(f"{len(tree.children)}: {tree.children[0]}")

    terminal: Tree = tree.children[1]

    if terminal.data == "primary":
        terminal_value: str = _process_primary(terminal)

    elif terminal.data == "expr":
        terminal_value = None
        while terminal and terminal.children:
            if terminal.data == "primary":
                terminal_value = _process_primary(terminal)
                break

            terminal: Tree = terminal.children[0]

        if not terminal_value:
            raise Exception(f"CAN NOT PROCESS MEMBER_INDEX terminal expr: {terminal}")

    else:
        raise Exception(f"UNKNOWN MEMBER_INDEX terminal TYPE: {terminal}")

    root: Tree = tree.children[0].children[0]

    if root.data == "member_dot":
        root_value = _process_member_dot(root)
    elif root.data == "member_index":
        root_value = _process_member_index(root)
    elif root.data == "member_dot_arg":
        root_value = _process_member_dot_arg(root)
    elif root.data == "primary":
        root_value: str = _process_primary(root)
    else:
        raise Exception(f"UNKNOWN MEMBER_INDEX root TYPE: {root}")

    return f"{root_value}.{terminal_value}"


def _process_primary(tree: Tree) -> str:
    if len(tree.children) != 1:
        raise Exception(f"UNKNOWN PRIMARY LENGTH! {len(tree.children)}: {tree}")

    primary: Tree = tree.children[0]

    # print(f"{len(tree.children)}: {primary}")

    if primary.data == "ident":
        return f"{primary.children[0]}"

    if primary.data == "literal":
        literal: Token = primary.children[0]
        if literal.type == "INT_LIT":
            return literal.value
        return literal.strip(literal[0])

    raise Exception(f"UNKNOWN PRIMARY DATA TYPE! {primary}")



from typing import NamedTuple

import celpy
from koreo.cel.prepare import Overlay


class ValueFunction(NamedTuple):
    preconditions: celpy.Runner | None
    local_values: celpy.Runner | None
    return_value: Overlay | None

    dynamic_input_keys: set[str]



import unittest

from celpy import celtypes
import celpy

from koreo.cel.functions import koreo_cel_functions, koreo_function_annotations


class TestToRef(unittest.TestCase):
    def test_invalid_type_to_ref(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = '"".to_ref()'
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        with self.assertRaises(celpy.CELEvalError):
            program.evaluate(inputs)

    def test_missing_name_and_external(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = (
            '{"apiVersion": "some.api.group/v1", "kind": "TestCase"}.to_ref()'
        )
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        with self.assertRaises(celpy.CELEvalError):
            program.evaluate(inputs)

    def test_to_ref_empty_name(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = '{"name": ""}.to_ref()'
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        with self.assertRaises(celpy.CELEvalError):
            program.evaluate(inputs)

    def test_to_ref_name_only(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = '{"name": "test-case"}.to_ref()'
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        result = program.evaluate(inputs)

        self.assertDictEqual({"name": "test-case"}, result)

    def test_to_ref_name_full(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = '{"apiVersion": "some.api.group/v2", "kind": "TestCase", "name": "test-case", "namespace": "a-namespace", "extra": "value"}.to_ref()'
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        result = program.evaluate(inputs)

        self.assertDictEqual(
            {
                "apiVersion": "some.api.group/v2",
                "kind": "TestCase",
                "name": "test-case",
                "namespace": "a-namespace",
            },
            result,
        )

    def test_to_ref_external_only(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = '{"external": "thing/some_id"}.to_ref()'
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        result = program.evaluate(inputs)

        self.assertDictEqual({"external": "thing/some_id"}, result)

    def test_to_ref_empty_external(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = '{"external": ""}.to_ref()'
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        with self.assertRaises(celpy.CELEvalError):
            program.evaluate(inputs)

    def test_to_ref_external_full(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = '{"apiVersion": "some.api.group/v2", "kind": "TestCase", "external": "group/id"}.to_ref()'
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        result = program.evaluate(inputs)

        self.assertDictEqual(
            {
                "apiVersion": "some.api.group/v2",
                "kind": "TestCase",
                "external": "group/id",
            },
            result,
        )


class TestSelfRef(unittest.TestCase):
    def test_invalid_resource_type(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = '"".self_ref()'
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        with self.assertRaises(celpy.CELEvalError):
            program.evaluate(inputs)

    def test_missing_api_version(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = '{"kind": "TestCase", "metadata": {"name": "a-name-value", "namespace": "some-namespace"}}.self_ref()'
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        with self.assertRaises(celpy.CELEvalError):
            program.evaluate(inputs)

    def test_missing_kind(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = '{"apiVersion": "some.api.group/v2", "metadata": {"name": "a-name-value", "namespace": "some-namespace"}}.self_ref()'
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        with self.assertRaises(celpy.CELEvalError):
            program.evaluate(inputs)

    def test_missing_metadata(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = (
            '{"apiVersion": "some.api.group/v2", "kind": "TestCase"}.self_ref()'
        )
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        with self.assertRaises(celpy.CELEvalError):
            program.evaluate(inputs)

    def test_missing_name(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = '{"apiVersion": "some.api.group/v2", "kind": "TestCase", "metadata": {"namespace": "some-namespace"}}.self_ref()'
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        with self.assertRaises(celpy.CELEvalError):
            program.evaluate(inputs)

    def test_missing_namespace(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = '{"apiVersion": "some.api.group/v2", "kind": "TestCase", "metadata": {"name": "a-name-value"}}.self_ref()'
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        with self.assertRaises(celpy.CELEvalError):
            program.evaluate(inputs)

    def test_to_ref_external_full(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = '{"apiVersion": "some.api.group/v2", "kind": "TestCase", "metadata": {"name": "a-name-value", "namespace": "some-namespace"}}.self_ref()'
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        result = program.evaluate(inputs)

        self.assertDictEqual(
            {
                "apiVersion": "some.api.group/v2",
                "kind": "TestCase",
                "name": "a-name-value",
                "namespace": "some-namespace",
            },
            result,
        )


class TestGroupRef(unittest.TestCase):
    def test_invalid_resource_type(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = '"".group_ref()'
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        with self.assertRaises(celpy.CELEvalError):
            program.evaluate(inputs)

    def test_missing_api_version(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = '{"kind": "TestCase", "metadata": {"name": "a-name-value", "namespace": "some-namespace"}}.group_ref()'
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        with self.assertRaises(celpy.CELEvalError):
            program.evaluate(inputs)

    def test_missing_kind(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = '{"apiVersion": "some.api.group/v2", "metadata": {"name": "a-name-value", "namespace": "some-namespace"}}.group_ref()'
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        with self.assertRaises(celpy.CELEvalError):
            program.evaluate(inputs)

    def test_missing_metadata(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = (
            '{"apiVersion": "some.api.group/v2", "kind": "TestCase"}.group_ref()'
        )
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        with self.assertRaises(celpy.CELEvalError):
            program.evaluate(inputs)

    def test_missing_name(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = '{"apiVersion": "some.api.group/v2", "kind": "TestCase", "metadata": {"namespace": "some-namespace"}}.group_ref()'
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        with self.assertRaises(celpy.CELEvalError):
            program.evaluate(inputs)

    def test_missing_namespace(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = '{"apiVersion": "some.api.group/v2", "kind": "TestCase", "metadata": {"name": "a-name-value"}}.group_ref()'
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        with self.assertRaises(celpy.CELEvalError):
            program.evaluate(inputs)

    def test_group_ref_from_api_version(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = '{"apiVersion": "some.api.group/v2", "kind": "TestCase", "name": "a-name-value", "namespace": "some-namespace"}.group_ref()'
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        result = program.evaluate(inputs)

        self.assertDictEqual(
            {
                "apiGroup": "some.api.group",
                "kind": "TestCase",
                "name": "a-name-value",
                "namespace": "some-namespace",
            },
            result,
        )

    def test_group_ref_from_api_group(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = '{"apiGroup": "some.api.group", "kind": "TestCase", "name": "a-name-value", "namespace": "some-namespace"}.group_ref()'
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        result = program.evaluate(inputs)

        self.assertDictEqual(
            {
                "apiGroup": "some.api.group",
                "kind": "TestCase",
                "name": "a-name-value",
                "namespace": "some-namespace",
            },
            result,
        )


class TestConfigConnectReady(unittest.TestCase):
    def test_invalid_resource_type(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = '"".config_connect_ready()'
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        result = program.evaluate(inputs)
        self.assertFalse(result)

    def test_no_status(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "{}.config_connect_ready()"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        result = program.evaluate(inputs)
        self.assertFalse(result)

    def test_no_conditions(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "{'status': {}}.config_connect_ready()"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        result = program.evaluate(inputs)
        self.assertFalse(result)

    def test_no_ready_condition(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "{'status': {'conditions': [{}]}}.config_connect_ready()"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        result = program.evaluate(inputs)
        self.assertFalse(result)

    def test_two_ready_conditions(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "{'status': {'conditions': [{'type': 'Ready'}, {'type': 'Ready'}]}}.config_connect_ready()"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        result = program.evaluate(inputs)
        self.assertFalse(result)

    def test_ready_condition_non_reason(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "{'status': {'conditions': [{'type': 'Ready', 'reason': 'Testing'}]}}.config_connect_ready()"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        result = program.evaluate(inputs)
        self.assertFalse(result)

    def test_non_ready_condition_non_status(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "{'status': {'conditions': [{'type': 'Ready', 'reason': 'UpToDate', 'status': 'False'}]}}.config_connect_ready()"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        result = program.evaluate(inputs)
        self.assertFalse(result)

    def test_ready_condition_true(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "{'status': {'conditions': [{'type': 'Ready', 'reason': 'UpToDate', 'status': 'True'}]}}.config_connect_ready()"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        result = program.evaluate(inputs)
        self.assertTrue(result)


class TestOverlay(unittest.TestCase):
    def test_empty_resource_and_overlay(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "{}.overlay({})"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        self.assertDictEqual({}, program.evaluate(inputs))

    def test_resource_and_empty_overlay(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = (
            "{'a': 'key', 'value': 18, 'bool': true, 'cel': 1 + 32}.overlay({})"
        )
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        self.assertDictEqual(
            {"a": "key", "value": 18, "bool": True, "cel": 33},
            program.evaluate(inputs),
        )

    def test_empty_resource_with_overlay(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = (
            "{}.overlay({'a': 'key', 'value': 18, 'bool': true, 'cel': 1 + 32})"
        )
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        self.assertDictEqual(
            {"a": "key", "value": 18, "bool": True, "cel": 33},
            program.evaluate(inputs),
        )

    def test_resource_with_overlay(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "{'a': 'wrong', 'value': 99, 'bool': false}.overlay({'a': 'key', 'value': 18, 'bool': true, 'cel': 1 + 32})"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        self.assertDictEqual(
            {"a": "key", "value": 18, "bool": True, "cel": 33},
            program.evaluate(inputs),
        )

    def test_resource_is_not_mutated(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "{'a': inputs.base.overlay({'value': inputs.base.value + 1}), 'b': inputs.base.overlay({'value': inputs.base.value + 1})}"
        inputs = {"inputs": celpy.json_to_cel({"base": {"value": 1}})}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)
        self.maxDiff = None
        self.assertDictEqual(
            {"a": {"value": 2}, "b": {"value": 2}},
            program.evaluate(inputs),
        )

    def test_resource_with_deep_overlay(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "{'nested': {'key': 'value'}}.overlay({'nested': {'key': 'value'}, 'new': {'nested': 'a' + string(1 + 8)}})"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        self.assertDictEqual(
            {"nested": {"key": "value"}, "new": {"nested": "a9"}},
            program.evaluate(inputs),
        )

    def test_resource_with_deep_update(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "{'nested': {'deep': {'key': 'value'}}}.overlay({'nested': { 'deep': { 'new_key': 'new_value', 'deeper': {'nested': true}}}})"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        self.assertDictEqual(
            {
                "nested": {
                    "deep": {
                        "key": "value",
                        "new_key": "new_value",
                        "deeper": {"nested": True},
                    },
                }
            },
            program.evaluate(inputs),
        )

    def test_resource_with_labels_overlay(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "{'metadata': {'labels': {'some.group/key': 'value'}, 'annotations': {'a.group/key': 'value'}}}.overlay({'metadata': {'labels': {'some.group/key': string(9 + 10 + 88)}, 'annotations': {'a.group/key': 'a ' + 'b ' + 'c'}}})"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        self.assertDictEqual(
            {
                "metadata": {
                    "labels": {"some.group/key": "107"},
                    "annotations": {"a.group/key": "a b c"},
                }
            },
            program.evaluate(inputs),
        )

    def test_overlay_with_dots_in_name(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "{'metadata': {'labels': {'some.group/key': 'value'}, 'annotations': {'a.group/key': 'value'}}}.overlay({'metadata.deep.value.name': 'a' + '-' + 'name'})"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        self.assertDictEqual(
            {
                "metadata": {
                    "labels": {"some.group/key": "value"},
                    "annotations": {"a.group/key": "value"},
                },
                "metadata.deep.value.name": "a-name",
            },
            program.evaluate(inputs),
        )


class TestFlatten(unittest.TestCase):
    def test_invalid_type(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "1.flatten()"

        with self.assertRaises(celpy.CELParseError):
            cel_env.compile(test_cel_expression)

    def test_empty_list(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "[].flatten()"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        flattened = program.evaluate(inputs)
        self.assertListEqual([], flattened)

    def test_flat(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "[1, 2, 3].flatten()"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        with self.assertRaises(celpy.CELEvalError):
            program.evaluate(inputs)

    def test_single(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "[[1, 2, 3]].flatten()"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        flattened = program.evaluate(inputs)
        self.assertListEqual([1, 2, 3], flattened)

    def test_multiple(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = '[[1, 2, 3], ["a", "b", "c"]].flatten()'
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        flattened = program.evaluate(inputs)
        self.assertListEqual([1, 2, 3, "a", "b", "c"], flattened)


class TestSplit(unittest.TestCase):
    def test_invalid_type(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "1.split('.')"

        with self.assertRaises(celpy.CELParseError):
            cel_env.compile(test_cel_expression)

    def test_empty_seperator(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "''.split('')"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        with self.assertRaises(celpy.CELEvalError):
            program.evaluate(inputs)

    def test_empty_target(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "''.split('.')"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        flattened = program.evaluate(inputs)
        self.assertListEqual([""], flattened)

    def test_no_spilt(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "'test-ing-is-fun'.split('.')"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        result = program.evaluate(inputs)
        self.assertListEqual(result, ["test-ing-is-fun"])

    def test_spilt(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "'test.ing.is.fun'.split('.')"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        result = program.evaluate(inputs)
        self.assertListEqual(result, ["test", "ing", "is", "fun"])


class TestSplitFirst(unittest.TestCase):
    def test_invalid_type(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "1.split_first('.')"

        with self.assertRaises(celpy.CELParseError):
            cel_env.compile(test_cel_expression)

    def test_empty_seperator(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "''.split_first('')"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        with self.assertRaises(celpy.CELEvalError):
            program.evaluate(inputs)

    def test_empty_target(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "''.split_first('.')"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        flattened = program.evaluate(inputs)
        self.assertEqual("", flattened)

    def test_no_spilt(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "'test-ing-is-fun'.split_first('.')"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        result = program.evaluate(inputs)
        self.assertEqual("test-ing-is-fun", result)

    def test_spilt(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "'test.ing.is.fun'.split_first('.')"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        result = program.evaluate(inputs)
        self.assertEqual("test", result)


class TestSplitIndex(unittest.TestCase):
    def test_invalid_type(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "1.split_index('.', 1)"

        with self.assertRaises(celpy.CELParseError):
            cel_env.compile(test_cel_expression)

    def test_empty_seperator(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "''.split_index('', 1)"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        with self.assertRaises(celpy.CELEvalError):
            program.evaluate(inputs)

    def test_empty_out_of_bounds(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "'test'.split_index('.', 2)"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        with self.assertRaises(celpy.CELEvalError):
            program.evaluate(inputs)

    def test_empty_target(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "''.split_index('.', 1)"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        flattened = program.evaluate(inputs)
        self.assertEqual("", flattened)

    def test_no_spilt(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "'test-ing-is-fun'.split_index('.', 0)"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        result = program.evaluate(inputs)
        self.assertEqual("test-ing-is-fun", result)

    def test_spilt(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "'test.ing.is.fun'.split_index('.', 2)"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        result = program.evaluate(inputs)
        self.assertEqual("is", result)


class TestSplitLast(unittest.TestCase):
    def test_invalid_type(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "1.split_last('.')"

        with self.assertRaises(celpy.CELParseError):
            cel_env.compile(test_cel_expression)

    def test_empty_seperator(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "''.split_last('')"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        with self.assertRaises(celpy.CELEvalError):
            program.evaluate(inputs)

    def test_empty_target(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "''.split_last('.')"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        flattened = program.evaluate(inputs)
        self.assertEqual("", flattened)

    def test_no_spilt(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "'test-ing-is-fun'.split_last('.')"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        result = program.evaluate(inputs)
        self.assertEqual("test-ing-is-fun", result)

    def test_spilt(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "'test.ing.is.fun'.split_last('.')"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        result = program.evaluate(inputs)
        self.assertEqual("fun", result)


class TestStrip(unittest.TestCase):
    def test_invalid_type(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "1.strip('test')"

        with self.assertRaises(celpy.CELParseError):
            cel_env.compile(test_cel_expression)

    def test_empty_on(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "test.strip('')"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        with self.assertRaises(celpy.CELEvalError):
            program.evaluate(inputs)

    def test_empty_target(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "''.strip('test')"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        flattened = program.evaluate(inputs)
        self.assertEqual("", flattened)

    def test_no_strip(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "'test-ing-is-fun'.strip('b')"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        result = program.evaluate(inputs)
        self.assertEqual("test-ing-is-fun", result)

    def test_strip(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "'testinga'.strip('testing')"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        result = program.evaluate(inputs)
        self.assertEqual("a", result)


class TestRStrip(unittest.TestCase):
    def test_invalid_type(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "1.rstrip('test')"

        with self.assertRaises(celpy.CELParseError):
            cel_env.compile(test_cel_expression)

    def test_empty_on(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "test.rstrip('')"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        with self.assertRaises(celpy.CELEvalError):
            program.evaluate(inputs)

    def test_empty_target(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "''.rstrip('test')"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        flattened = program.evaluate(inputs)
        self.assertEqual("", flattened)

    def test_no_rstrip(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "'test-ing-is-fun'.rstrip('b')"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        result = program.evaluate(inputs)
        self.assertEqual("test-ing-is-fun", result)

    def test_rstrip(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "'testingisfun'.rstrip('isfun')"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        result = program.evaluate(inputs)
        self.assertEqual("testing", result)


class TestToJson(unittest.TestCase):
    def test_to_json_null(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "to_json()"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        with self.assertRaises(celpy.CELEvalError):
            program.evaluate(inputs)

    def test_to_json_str(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "to_json('isfun')"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        result = program.evaluate(inputs)
        self.assertEqual('"isfun"', result)

    def test_to_json_map(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = 'to_json({"testKey": "testValue"})'
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        result = program.evaluate(inputs)
        self.assertEqual('{"testKey": "testValue"}', result)

    def test_to_json_list(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = 'to_json(["testKey", "testValue"])'
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        result = program.evaluate(inputs)
        self.assertEqual('["testKey", "testValue"]', result)

    def test_to_json_bool(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "to_json(false)"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        result = program.evaluate(inputs)
        self.assertEqual("0", result)

    def test_to_json_int(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "to_json(3)"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        result = program.evaluate(inputs)
        self.assertEqual("3", result)


class TestFromJson(unittest.TestCase):
    def test_from_json_null(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "from_json()"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        with self.assertRaises(celpy.CELEvalError):
            program.evaluate(inputs)

    def test_from_json_str(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "from_json('isfun')"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        with self.assertRaises(celpy.CELEvalError):
            program.evaluate(inputs)

    def test_from_json_map(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = 'from_json(\'{"testKey": "testValue"}\')'
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        result = program.evaluate(inputs)
        self.assertEqual({"testKey": "testValue"}, result)

    def test_from_json_list(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = 'from_json(\'["testKey", "testValue"]\')'
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        result = program.evaluate(inputs)
        self.assertEqual(["testKey", "testValue"], result)

    def test_from_json_bool(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "from_json(false)"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        with self.assertRaises(celpy.CELEvalError):
            program.evaluate(inputs)

    def test_from_json_int(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "from_json(3)"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        with self.assertRaises(celpy.CELEvalError):
            program.evaluate(inputs)


class TestBase64Encode(unittest.TestCase):
    def test_b64encode_null(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "b64encode()"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        with self.assertRaises(celpy.CELEvalError):
            program.evaluate(inputs)

    def test_b64encode_str(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "b64encode('isfun')"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        result = program.evaluate(inputs)
        self.assertEqual("aXNmdW4=", result)

    def test_b64encode_map(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = 'b64encode({"testKey": "testValue"})'
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        with self.assertRaises(celpy.CELEvalError):
            program.evaluate(inputs)

    def test_b64encode_list(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = 'b64encode(["testKey", "testValue"])'
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        with self.assertRaises(celpy.CELEvalError):
            program.evaluate(inputs)

    def test_b64encode_bool(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "b64encode(false)"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        with self.assertRaises(celpy.CELEvalError):
            program.evaluate(inputs)

    def test_b64encode_int(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "b64encode(3)"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        with self.assertRaises(celpy.CELEvalError):
            program.evaluate(inputs)


class TestBase64Decode(unittest.TestCase):
    def test_b64decode_null(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "b64decode()"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        with self.assertRaises(celpy.CELEvalError):
            program.evaluate(inputs)

    def test_b64decode_str(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "b64decode('aXNmdW4=')"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        result = program.evaluate(inputs)
        self.assertEqual("isfun", result)

    def test_b64encode_invalid(self):
        cel_env = celpy.Environment(annotations=koreo_function_annotations)

        test_cel_expression = "b64decode('aXNmdW')"
        inputs = {}

        compiled = cel_env.compile(test_cel_expression)
        program = cel_env.program(compiled, functions=koreo_cel_functions)

        with self.assertRaises(celpy.CELEvalError):
            program.evaluate(inputs)



import json
import unittest

from celpy import celtypes

from koreo.cel.encoder import convert_bools, encode_cel


class TestEncodeBools(unittest.TestCase):
    def test_structure(self):
        value = {
            "zero": 0,
            "one": 1,
            celtypes.BoolType(True): celtypes.BoolType(True),
            celtypes.StringType("true"): celtypes.BoolType(True),
            celtypes.BoolType(False): celtypes.BoolType(False),
            celtypes.StringType("false"): celtypes.BoolType(False),
            "nested_map": {
                "zero": 0,
                "one": 1,
                celtypes.BoolType(True): celtypes.BoolType(True),
                celtypes.StringType("true"): celtypes.BoolType(True),
                celtypes.BoolType(False): celtypes.BoolType(False),
                celtypes.StringType("false"): celtypes.BoolType(False),
            },
            "list": [
                "zero",
                0,
                "one",
                1,
                celtypes.BoolType(True),
                celtypes.StringType("true"),
                celtypes.BoolType(False),
                celtypes.StringType("false"),
            ],
            "cel_values": celtypes.MapType(
                {
                    celtypes.StringType("zero"): celtypes.IntType(0),
                    celtypes.StringType("one"): celtypes.IntType(1),
                    celtypes.BoolType(True): celtypes.BoolType(True),
                    celtypes.StringType("true"): celtypes.BoolType(True),
                    celtypes.StringType("false"): celtypes.BoolType(False),
                    celtypes.StringType("nested_map"): celtypes.MapType(
                        {
                            celtypes.StringType("zero"): celtypes.IntType(0),
                            celtypes.StringType("one"): celtypes.IntType(1),
                            celtypes.BoolType(True): celtypes.BoolType(True),
                            celtypes.StringType("true"): celtypes.BoolType(True),
                            celtypes.BoolType(False): celtypes.BoolType(False),
                            celtypes.StringType("false"): celtypes.BoolType(False),
                        }
                    ),
                    celtypes.StringType("list"): celtypes.ListType(
                        [
                            celtypes.StringType("zero"),
                            celtypes.IntType(0),
                            celtypes.StringType("one"),
                            celtypes.IntType(1),
                            celtypes.BoolType(True),
                            celtypes.StringType("true"),
                            celtypes.BoolType(False),
                            celtypes.StringType("false"),
                        ]
                    ),
                }
            ),
        }
        self.maxDiff = None
        self.assertEqual(
            '{"zero": 0, "one": 1, "true": true, "true": true, "false": false, "false": false, "nested_map": {"zero": 0, "one": 1, "true": true, "true": true, "false": false, "false": false}, "list": ["zero", 0, "one", 1, true, "true", false, "false"], "cel_values": {"zero": 0, "one": 1, "true": true, "true": true, "false": false, "nested_map": {"zero": 0, "one": 1, "true": true, "true": true, "false": false, "false": false}, "list": ["zero", 0, "one", 1, true, "true", false, "false"]}}',
            json.dumps(convert_bools(value)),
        )


class TestEncodeCel(unittest.TestCase):
    def test_empty_str(self):
        value = ""
        self.assertEqual('""', encode_cel(value))

    def test_str(self):
        value = "This is a plain old string."
        self.assertEqual('"This is a plain old string."', encode_cel(value))

    def test_str_with_quote(self):
        value = 'This is a "test."'
        self.assertEqual('"""This is a \\\"test.\\\""""', encode_cel(value))  # fmt: skip

    def test_str_with_tripple_quote(self):
        value = 'This is a """test."""'
        self.assertEqual('"""This is a \\\"\\\"\\\"test.\\\"\\\"\\\""""', encode_cel(value))  # fmt: skip

    def test_expression_str(self):
        value = "=1 + 1"
        self.assertEqual("1 + 1", encode_cel(value))

    def test_int(self):
        value = 1010
        self.assertEqual("1010", encode_cel(value))

    def test_int_str(self):
        value = "3213"
        self.assertEqual("3213", encode_cel(value))

    def test_float(self):
        value = 99.3
        self.assertEqual("99.3", encode_cel(value))

    def test_float_str(self):
        value = "72.3"
        self.assertEqual("72.3", encode_cel(value))

    def test_bool(self):
        value = True
        self.assertEqual("true", encode_cel(value))

    def test_list_of_int(self):
        value = [1, 2, 3]
        self.assertEqual("[1,2,3]", encode_cel(value))

    def test_list_of_float(self):
        value = [1.2, 2.1, 8.9, 9.0]
        self.assertEqual("[1.2,2.1,8.9,9.0]", encode_cel(value))

    def test_list_of_bool(self):
        value = [False, True, True, False]
        self.assertEqual("[false,true,true,false]", encode_cel(value))

    def test_flat_dict(self):
        value = {
            "a_string": "testing",
            "a_quoted_string": 'you should "test"',
            "a_tripple_quoted_string": 'you """should""" "test"',
            "an_int": 7,
            "an_int_str": "29",
            "a_float": 82.34,
            "a_float_str": "94.55",
            "bool_true": True,
            "bool_false": False,
            "empty_list": [],
            "complex_list": ["a", 2, "4", 3.2, "53.4", True, False],
            "nested_list": [[1, 2, 3], [True, False, False], ["a", "b", "c"]],
            "cel_expr": "=8 + 3",
            "cel_expr_list": ["=8 + 3", "='a' + 'b'"],
            "cel_expr_list_list": [["=8 + 3", "='a' + 'b'"], ["=has(value)"]],
            "false": False,
            "true": True,
            "yes": True,
            "no": False,
            "none": None,
            "null": "=null",
        }

        # fmt: off
        self.maxDiff = None
        self.assertEqual(
            '{"a_string":"testing","a_quoted_string":"""you should \\\"test\\\"""","a_tripple_quoted_string":"""you \\\"\\\"\\\"should\\\"\\\"\\\" \\\"test\\\"""","an_int":7,"an_int_str":29,"a_float":82.34,"a_float_str":94.55,"bool_true":true,"bool_false":false,"empty_list":[],"complex_list":["a",2,4,3.2,53.4,true,false],"nested_list":[[1,2,3],[true,false,false],["a","b","c"]],"cel_expr":8 + 3,"cel_expr_list":[8 + 3,\'a\' + \'b\'],"cel_expr_list_list":[[8 + 3,\'a\' + \'b\'],[has(value)]],"false":false,"true":true,"yes":true,"no":false,"none":null,"null":null}',
            encode_cel(value),
        )
        # fmt: on

    def test_nested_dict(self):
        value = {
            "strings": {
                "a_string": "testing",
                "a_quoted_string": '"testing" is important',
                "an_int_str": "29",
                "a_float_str": "94.55",
            },
            "numbers": {"an_int": 7, "a_float": 82.34},
            "bools": {"bool_true": True, "bool_false": False},
            "lists": {
                "empty_list": [],
                "complex_list": ["a", 2, "4", 3.2, "53.4", True, False],
                "nested_list": [[1, 2, 3], [True, False, False], ["a", "b", "c"]],
            },
            "cel": {
                "cel_expr": "=8 + 3",
                "cel_expr_list": ["=8 + 3", "='a' + 'b'"],
                "cel_expr_list_list": [["=8 + 3", "='a' + 'b'"], ["=has(value)"]],
            },
            "dicts": {
                "level_one": {
                    "level_two": {"value": "deep", "cel": "=has(formula)"},
                    "sibling": "=8 + 99",
                },
                "sibling": "a",
            },
        }
        self.maxDiff = None
        # fmt: off
        self.assertEqual(
            '{"strings":{"a_string":"testing","a_quoted_string":"""\\\"testing\\\" is important""","an_int_str":29,"a_float_str":94.55},"numbers":{"an_int":7,"a_float":82.34},"bools":{"bool_true":true,"bool_false":false},"lists":{"empty_list":[],"complex_list":["a",2,4,3.2,53.4,true,false],"nested_list":[[1,2,3],[true,false,false],["a","b","c"]]},"cel":{"cel_expr":8 + 3,"cel_expr_list":[8 + 3,\'a\' + \'b\'],"cel_expr_list_list":[[8 + 3,\'a\' + \'b\'],[has(value)]]},"dicts":{"level_one":{"level_two":{"value":"deep","cel":has(formula)},"sibling":8 + 99},"sibling":"a"}}',
            encode_cel(value),
        )
        # fmt: on



from collections import defaultdict


__workflow_custom_crd_index = defaultdict(str)
__custom_crd_wokflow_index = defaultdict(set[str])


def index_workflow_custom_crd(workflow: str, custom_crd: str):
    prior_custom_crd = __workflow_custom_crd_index[workflow]

    if prior_custom_crd == custom_crd:
        return

    if workflow in __custom_crd_wokflow_index[prior_custom_crd]:
        __custom_crd_wokflow_index[prior_custom_crd].remove(workflow)

    __custom_crd_wokflow_index[custom_crd].add(workflow)

    __workflow_custom_crd_index[workflow] = custom_crd


def unindex_workflow_custom_crd(workflow: str):
    prior_custom_crd = __workflow_custom_crd_index[workflow]

    if not prior_custom_crd:
        return

    if workflow in __custom_crd_wokflow_index[prior_custom_crd]:
        __custom_crd_wokflow_index[prior_custom_crd].remove(workflow)

    del __workflow_custom_crd_index[workflow]


def get_custom_crd_workflows(custom_crd: str) -> list[str]:
    return list(__custom_crd_wokflow_index[custom_crd])


def _reset_registry():
    global __workflow_custom_crd_index, __custom_crd_wokflow_index
    __workflow_custom_crd_index = defaultdict(str)
    __custom_crd_wokflow_index = defaultdict(set[str])



from typing import Any, NamedTuple
import logging

import celpy

from koreo.cel.encoder import encode_cel
from koreo.cel.functions import koreo_cel_functions
from koreo.result import PermFail


def prepare_expression(
    cel_env: celpy.Environment, spec: Any | None, location: str
) -> None | celpy.Runner | PermFail:
    if not spec:
        return None

    try:
        encoded = encode_cel(spec)
    except Exception as err:
        return PermFail(
            message=f"Structural error in {location}, while building expression '{err}'.",
        )

    try:
        value = cel_env.program(cel_env.compile(encoded), functions=koreo_cel_functions)
    except celpy.CELParseError as err:
        return PermFail(
            message=f"Parsing error at line {err.line}, column {err.column}. '{err}' in {location} ('{encoded}')",
        )

    value.logger.setLevel(logging.WARNING)

    return value


def prepare_map_expression(
    cel_env: celpy.Environment, spec: Any | None, location: str
) -> None | celpy.Runner | PermFail:
    if not spec:
        return None

    if not isinstance(spec, dict):
        return PermFail(message=f"Malformed {location}, expected a mapping")

    return prepare_expression(cel_env=cel_env, spec=spec, location=location)


Index = dict[str, "Index"] | int


class Overlay(NamedTuple):
    value_index: Index
    values: celpy.Runner


def prepare_overlay_expression(
    cel_env: celpy.Environment, spec: Any | None, location: str
) -> None | Overlay | PermFail:
    if not spec:
        return None

    if not isinstance(spec, dict):
        return PermFail(message=f"Malformed {location}, expected a mapping")

    overlay_index, overlay_values = _overlay_indexer(spec=spec, base=0)

    match prepare_expression(cel_env=cel_env, spec=overlay_values, location=location):
        case None:
            return None
        case PermFail() as err:
            return err
        case celpy.Runner() as value_expression:
            return Overlay(value_index=overlay_index, values=value_expression)


def _overlay_indexer(spec: dict, base: int = 0) -> tuple[Index, list]:
    index = {}
    values = []
    for key, value in spec.items():
        match value:
            case dict() if value:
                index[key], key_values = _overlay_indexer(
                    value, base=len(values) + base
                )
                values.extend(key_values)
            case _:
                index[key] = len(values) + base
                values.append(value)

    return index, values



from typing import Any
import logging
import pathlib

logger = logging.getLogger("koreo.schema")

import fastjsonschema
import yaml

from koreo.function_test.structure import FunctionTest
from koreo.resource_function.structure import ResourceFunction
from koreo.resource_template.structure import ResourceTemplate
from koreo.result import PermFail
from koreo.value_function.structure import ValueFunction
from koreo.workflow.structure import Workflow

DEFAULT_API_VERSION = "v1beta1"

CRD_ROOT = pathlib.Path(__file__).parent.parent.parent.joinpath("crd")

CRD_MAP = {
    FunctionTest: CRD_ROOT.joinpath("function-test.yaml"),
    ResourceFunction: CRD_ROOT.joinpath("resource-function.yaml"),
    ResourceTemplate: CRD_ROOT.joinpath("resource-template.yaml"),
    ValueFunction: CRD_ROOT.joinpath("value-function.yaml"),
    Workflow: CRD_ROOT.joinpath("workflow.yaml"),
}

_SCHEMA_VALIDATORS = {}


def validate(
    resource_type: type,
    spec: Any,
    schema_version: str | None = None,
    validation_required: bool = False,
):
    schema_validator = _get_validator(
        resource_type=resource_type, version=schema_version
    )
    if not schema_validator:
        if not validation_required:
            return None

        return PermFail(
            f"Schema validator not found for {resource_type.__name__} version {schema_version or DEFAULT_API_VERSION}",
        )

    try:
        schema_validator(spec)
    except fastjsonschema.JsonSchemaValueException as err:
        # This is hacky, and likely buggy, but it makes the messages easier to grok.
        validation_err = f"{err.rule_definition} {err}".replace(
            "data.", "spec."
        ).replace("data ", "spec ")
        return PermFail(validation_err)

    return None


def _get_validator(resource_type: type, version: str | None = None):
    if not version:
        version = DEFAULT_API_VERSION

    resource_version_key = f"{resource_type.__qualname__}:{version}"

    return _SCHEMA_VALIDATORS.get(resource_version_key)


def _load_validators():
    _SCHEMA_VALIDATORS.clear()

    for resource_type, schema_file in CRD_MAP.items():
        resource_validators = _load_validator(schema_file)
        if not resource_validators:
            continue

        for version, validator in resource_validators:
            resource_version_key = f"{resource_type.__qualname__}:{version}"
            _SCHEMA_VALIDATORS[resource_version_key] = validator


def _load_validator(schema_file: pathlib.Path):

    if not schema_file.exists():
        return None

    with schema_file.open() as crd_content:
        parsed = yaml.load(crd_content, Loader=yaml.Loader)
        spec = parsed.get("spec")
        if not spec:
            return None

        spec_names = spec.get("names")
        if spec_names:
            spec_kind = spec_names.get("kind", "<missing kind>")
        else:
            spec_kind = "<missing kind>"

        schema_specs = spec.get("versions")
        if not schema_specs:
            return None

        validators = []

        for schema_spec in schema_specs:
            version = schema_spec.get("name")
            if not version:
                continue

            schema_block = schema_spec.get("schema")
            if not schema_block:
                continue

            openapi_schema = schema_block.get("openAPIV3Schema")
            if not openapi_schema:
                continue

            openapi_properties = openapi_schema.get("properties")
            if not openapi_properties:
                continue

            openapi_spec = openapi_properties.get("spec")

            try:
                validators.append((version, fastjsonschema.compile(openapi_spec)))
            except fastjsonschema.JsonSchemaDefinitionException:
                logger.exception(f"Failed to process {spec_kind} {version}")
                pass
            except AttributeError as err:
                logger.error(
                    f"Probably encountered an empty `properties` block for {spec_kind} {version} (err: {err})"
                )
                raise

        return validators


_load_validators()



from datetime import UTC, datetime
from unittest.mock import patch
import unittest

from resources.k8s.conditions import Condition, Conditions, update_condition


class TestConditions(unittest.TestCase):
    def test_new(self):
        conditions: Conditions = []
        update: Condition = Condition(
            type="NewTest",
            reason="Testing",
            status="AwaitDependency",
            message="Testing new condition insertion.",
            location="testing.location",
        )

        frozen_now = datetime.now(UTC)
        with patch("resources.k8s.conditions.datetime") as frozen_now_datetime:
            frozen_now_datetime.now.return_value = frozen_now
            updated_conditions = update_condition(
                conditions=conditions, condition=update
            )

        self.assertFalse(conditions)
        self.assertEqual(1, len(updated_conditions))

        condition = updated_conditions[0]
        self.assertDictEqual(
            condition,
            {
                "lastUpdateTime": frozen_now.isoformat(),
                "lastTransitionTime": frozen_now.isoformat(),
                "type": "NewTest",
                "reason": "Testing",
                "status": "AwaitDependency",
                "message": "Testing new condition insertion.",
                "location": "testing.location",
            },
        )

    def test_non_update_update(self):
        conditions: Conditions = []
        condition: Condition = Condition(
            type="UpdateTest",
            reason="Done",
            status="Ready",
            message="All set.",
            location="testing.location",
        )

        insert_frozen_now = datetime.now(UTC)
        with patch("resources.k8s.conditions.datetime") as frozen_now_datetime:
            frozen_now_datetime.now.return_value = insert_frozen_now

            updated_conditions = update_condition(
                conditions=conditions, condition=condition
            )

        update_frozen_now = datetime.now(UTC)
        with patch("resources.k8s.conditions.datetime") as frozen_now_datetime:
            frozen_now_datetime.now.return_value = update_frozen_now

            updated_conditions = update_condition(
                conditions=updated_conditions, condition=condition
            )

        self.assertFalse(conditions)
        self.assertEqual(1, len(updated_conditions))

        self.assertDictEqual(
            updated_conditions[0],
            {
                "lastUpdateTime": update_frozen_now.isoformat(),
                "lastTransitionTime": insert_frozen_now.isoformat(),
                "type": "UpdateTest",
                "reason": "Done",
                "status": "Ready",
                "message": "All set.",
                "location": "testing.location",
            },
        )

    def test_multiple_conditions(self):
        conditions: Conditions = []
        update_one: Condition = Condition(
            type="UpdateTest",
            reason="Testing",
            status="AwaitDependency",
            message="Testing new condition insertion.",
            location="testing.location.1",
        )

        updated_conditions = update_condition(
            conditions=conditions, condition=update_one
        )

        new: Condition = Condition(
            type="NewTest",
            reason="Waiting",
            status="AwaitDependency",
            message="Testing new condition insertion.",
            location="testing.location.new",
        )

        new_frozen_now = datetime.now(UTC)
        with patch("resources.k8s.conditions.datetime") as frozen_now_datetime:
            frozen_now_datetime.now.return_value = new_frozen_now

            updated_conditions = update_condition(
                conditions=updated_conditions, condition=new
            )

        update_two: Condition = Condition(
            type="UpdateTest",
            reason="Done",
            status="Ready",
            message="All set.",
            location="testing.location.2",
        )

        update_frozen_now = datetime.now(UTC)
        with patch("resources.k8s.conditions.datetime") as frozen_now_datetime:
            frozen_now_datetime.now.return_value = update_frozen_now

            updated_conditions = update_condition(
                conditions=updated_conditions, condition=update_two
            )

        self.assertFalse(conditions)
        self.assertEqual(2, len(updated_conditions))

        # NOTE: I do not like how this assumes a specific order of conditions
        self.assertDictEqual(
            updated_conditions[0],
            {
                "lastUpdateTime": update_frozen_now.isoformat(),
                "lastTransitionTime": update_frozen_now.isoformat(),
                "type": "UpdateTest",
                "reason": "Done",
                "status": "Ready",
                "message": "All set.",
                "location": "testing.location.2",
            },
        )

        self.assertDictEqual(
            updated_conditions[1],
            {
                "lastUpdateTime": new_frozen_now.isoformat(),
                "lastTransitionTime": new_frozen_now.isoformat(),
                "type": "NewTest",
                "reason": "Waiting",
                "status": "AwaitDependency",
                "message": "Testing new condition insertion.",
                "location": "testing.location.new",
            },
        )



from __future__ import annotations

from functools import reduce
from typing import Any, NoReturn, TypeIs, TypeVar, Iterable

import kopf


class DepSkip:
    """This is internal for skipping due to dependency fail."""

    message: str | None
    location: str | None

    def __init__(self, message: str | None = None, location: str | None = None):
        self.message = message
        self.location = location

    def combine(self, other: Outcome):
        return other

    def __str__(self) -> str:
        if self.message:
            return f"Dependency Skip (message={self.message})"

        return "Dependency Violation Skip"


class Skip:
    """Indicates that this was intentially skipped."""

    message: str | None
    location: str | None

    def __init__(self, message: str | None = None, location: str | None = None):
        self.message = message
        self.location = location

    def combine(self, other: Outcome):
        if isinstance(other, (DepSkip,)):
            return self

        return other

    def __str__(self) -> str:
        if self.message:
            return f"Skip (message={self.message})"

        return "User Skip"


class Ok[T]:
    """Indicates success and `self.data` contains a value of type `T`."""

    data: T
    location: str | None

    def __init__(self, data: T, location: str | None = None):
        self.data = data
        self.location = location

    def __str__(self) -> str:
        return f"Ok(data={self.data})"

    def combine(self, other: Outcome):
        # This is a lower-priority outcome
        if isinstance(other, (DepSkip, Skip)):
            if isinstance(self.data, _OkData):
                return self

            return Ok(data=_OkData([self.data]), location=self.location)

        if not isinstance(other, Ok):
            return other

        if isinstance(self.data, _OkData):
            combined_data = self.data.append(other.data)
        else:
            combined_data = _OkData([self.data, other.data])

        location = []
        if self.location:
            location.append(self.location)

        if other.location:
            location.append(other.location)

        return Ok(data=combined_data, location=", ".join(location))


class Retry:
    """Retry reconciliation after `self.delay` seconds."""

    message: str | None
    delay: int
    location: str | None

    def __init__(
        self, delay: int = 60, message: str | None = None, location: str | None = None
    ):
        self.message = message
        self.delay = delay
        self.location = location

    def __str__(self) -> str:
        return f"Retry(delay={self.delay}, message={self.message})"

    def combine(self, other: Outcome):
        if isinstance(other, (DepSkip, Skip, Ok)):
            return self

        if isinstance(other, PermFail):
            return other

        message = []
        if self.message:
            message.append(self.message)

        if other.message:
            message.append(other.message)

        location = []
        if self.location:
            location.append(self.location)

        if other.location:
            location.append(other.location)

        return Retry(
            message=", ".join(message),
            delay=max(self.delay, other.delay),
            location=", ".join(location),
        )


class PermFail:
    """An error indicating retries should not be attempted."""

    message: str | None
    location: str | None

    def __init__(self, message: str | None = None, location: str | None = None):
        self.message = message
        self.location = location

    def __str__(self) -> str:
        return f"Permanent Failure (message={self.message})"

    def combine(self, other: Outcome):
        if isinstance(other, (DepSkip, Skip, Ok, Retry)):
            return self

        message = []
        if self.message:
            message.append(self.message)

        if other.message:
            message.append(other.message)

        location = []
        if self.location:
            location.append(self.location)

        if other.location:
            location.append(other.location)

        return PermFail(message=", ".join(message), location=", ".join(location))


OkT = TypeVar("OkT")

ErrorOutcome = Retry | PermFail
SkipOutcome = DepSkip | Skip
NonOkOutcome = SkipOutcome | ErrorOutcome
Outcome = NonOkOutcome | Ok[OkT]
UnwrappedOutcome = NonOkOutcome | OkT


# This is just for the Ok combine's use so that we know if we've already
# combined results.
class _OkData:
    def __init__(self, values: list):
        self.values = values

    def append(self, value):
        new = self.values[:]
        new.append(value)
        return _OkData(new)


def combine(outcomes: Iterable[Outcome]):
    if not outcomes:
        return Skip()

    combined = reduce(lambda acc, outcome: acc.combine(outcome), outcomes, DepSkip())

    if not isinstance(combined, Ok):
        return combined

    if isinstance(combined.data, _OkData):
        return Ok(data=combined.data.values, location=combined.location)

    # Python will not run reduce if there's a single element in the list, it
    # returns the first element, so the value should be wrapped.
    return Ok(data=[combined.data], location=combined.location)


def unwrapped_combine(outcomes: Iterable[UnwrappedOutcome]) -> UnwrappedOutcome:
    if not outcomes:
        return Skip()

    combined = reduce(
        lambda acc, outcome: acc.combine(
            Ok(outcome) if is_unwrapped_ok(outcome) else outcome
        ),
        outcomes,
        DepSkip(),  # The lowest priority in a combine flow
    )

    if not is_ok(combined):
        return combined

    if isinstance(combined.data, _OkData):
        return combined.data.values

    # Python will not run reduce if there's a single element in the list, it
    # returns the first element, so the value should be wrapped.
    return [combined.data]


def is_ok[T](candidate: Outcome[T]) -> TypeIs[Ok[T]]:
    if isinstance(candidate, Ok):
        return True

    return False


def is_unwrapped_ok[T](candidate: UnwrappedOutcome[T]) -> TypeIs[T]:
    if is_error(candidate):
        return False

    if is_skip(candidate):
        return False

    return True


def is_not_ok(candidate: Outcome) -> TypeIs[NonOkOutcome]:
    return not is_ok(candidate=candidate)


def is_error(candidate: Any) -> TypeIs[ErrorOutcome]:
    if isinstance(candidate, (Retry, PermFail)):
        return True

    return False


def is_skip(candidate: Any) -> TypeIs[SkipOutcome]:
    if isinstance(candidate, (DepSkip, Skip)):
        return True

    return False


def is_not_error(candidate: Outcome) -> TypeIs[Ok | Skip | DepSkip]:
    return not is_error(candidate=candidate)


def raise_for_error(error: Retry | PermFail) -> NoReturn:
    if isinstance(error, Retry):
        raise kopf.TemporaryError(error.message, delay=error.delay)

    raise kopf.PermanentError(error.message)



from __future__ import annotations
from contextlib import asynccontextmanager
from typing import NamedTuple, Sequence
import copy
import json

import kr8s

import celpy
from celpy import celtypes

from constants import LAST_APPLIED_ANNOTATION

from koreo import result
from koreo.cel.encoder import convert_bools
from koreo.cel.evaluation import evaluate_overlay
from koreo.cel.functions import _overlay

from koreo.resource_function.reconcile import reconcile_resource_function
from koreo.resource_function.structure import ResourceFunction
from koreo.value_function.reconcile import reconcile_value_function
from koreo.value_function.structure import ValueFunction

from .structure import (
    ExpectDelete,
    ExpectOutcome,
    ExpectResource,
    ExpectReturn,
    FunctionTest,
    TestCase,
)


class MockResponse:
    def __init__(self, data):
        self._data = data

    def json(self):
        return self._data


class MockApi:
    def __init__(self, current_resource: dict | None, *args, **kwargs):
        self._current_resource = current_resource
        self._materialized = None
        self._api_called = False
        self._delete_called = False

    @property
    def materialized(self):
        return self._materialized

    @property
    def namespace(self):
        return "FAKE-NAMESPACE"

    async def async_get(self, *args, **kwargs):
        if self._current_resource:
            # TODO: This should probably be loaded and built from the Function,
            # using _build_resource_config

            resource_class = kr8s.objects.new_class(
                version=self._current_resource.get("apiVersion"),
                kind=self._current_resource.get("kind"),
                namespaced=(
                    True
                    if self._current_resource.get("metadata", {}).get("namespace")
                    else False
                ),
                asyncio=True,
            )
            yield resource_class(
                api=self,
                resource=self._current_resource,
                namespace=self._current_resource.get("metadata", {}).get("namespace"),
            )

    @asynccontextmanager
    async def call_api(self, *args, **kwargs):
        self._api_called = True

        if args and "DELETE" in args:
            self._materialized = {}
            self._delete_called = True
            yield MockResponse(data=None)
            return

        data = json.loads(kwargs.get("data", "{}"))

        if not self._current_resource:
            merged = data
        else:
            merged = _merge_overlay(self._current_resource, data)

        self._materialized = merged
        yield MockResponse(data=merged)


def _merge_overlay(base, overlay):
    updated = copy.deepcopy(base)
    for key, value in overlay.items():
        match value:
            case dict():
                match base.get(key):
                    case dict() as sub_base:
                        updated[key] = _merge_overlay(sub_base, value)

        updated[key] = value

    return updated


class TestCaseResult(NamedTuple):
    test_pass: bool

    expected_outcome: result.Outcome | None = None
    outcome: result.UnwrappedOutcome | None = None

    message: str | None = None
    differences: Difference | None = None

    label: str | None = None


class TestResult(NamedTuple):
    test_results: Sequence[TestCaseResult]
    function_healthy: result.UnwrappedOutcome
    fatal_error: bool = False


class TestCaseOutcome(NamedTuple):
    new_inputs: celtypes.MapType | None
    new_resource: dict | None
    result: TestCaseResult
    fatal_error: bool = False


async def run_function_test(location: str, function_test: FunctionTest) -> TestResult:
    if not result.is_unwrapped_ok(function_test.function_under_test):
        return TestResult(
            test_results=[],
            function_healthy=function_test.function_under_test,
            fatal_error=True,
        )

    if not function_test.test_cases:
        return TestResult(test_results=[], function_healthy=result.Ok(None))

    test_results, fatal_error = await _run_test_cases(
        location=location,
        function_under_test=function_test.function_under_test,
        base_inputs=function_test.inputs,
        initial_resource=function_test.initial_resource,
        test_cases=function_test.test_cases,
    )
    return TestResult(
        test_results=test_results,
        function_healthy=result.Ok(None),
        fatal_error=fatal_error,
    )


async def _run_test_cases(
    location: str,
    function_under_test: ResourceFunction | ValueFunction,
    base_inputs: celtypes.MapType | None,
    initial_resource: dict | None,
    test_cases: Sequence[TestCase],
) -> tuple[Sequence[TestCaseResult], bool]:
    current_inputs = base_inputs
    current_resource = initial_resource

    test_results: list[TestCaseResult] = []
    for idx, test_case in enumerate(test_cases):
        result = await _run_test_case(
            location=location,
            current_resource=copy.deepcopy(current_resource),
            function_under_test=function_under_test,
            base_inputs=copy.deepcopy(current_inputs),
            test_case=test_case,
            idx=idx,
        )

        current_resource = result.new_resource
        current_inputs = result.new_inputs

        test_results.append(result.result)

        if result.fatal_error:
            return test_results, True

    return test_results, False


async def _run_test_case(
    location: str,
    function_under_test: ResourceFunction | ValueFunction,
    current_resource: dict | None,
    base_inputs: celtypes.MapType | None,
    test_case: TestCase,
    idx: int,
) -> TestCaseOutcome:
    location = f"{location}.testCases[{idx}]"
    if test_case.skip:
        return TestCaseOutcome(
            new_inputs=base_inputs,
            new_resource=current_resource,
            result=TestCaseResult(
                test_pass=True,
                expected_outcome=None,
                outcome=None,
                message=f"user skipped",
                label=test_case.label,
            ),
            fatal_error=False,
        )

    if base_inputs and test_case.input_overrides:
        inputs = _overlay(base_inputs, test_case.input_overrides)
        if isinstance(inputs, celpy.CELEvalError):
            return TestCaseOutcome(
                new_inputs=base_inputs,
                new_resource=current_resource,
                result=TestCaseResult(
                    test_pass=False,
                    expected_outcome=None,
                    outcome=None,
                    message=f"Could not run test case due to inputs overlay error: ({inputs})",
                    label=test_case.label,
                ),
                fatal_error=not test_case.variant,
            )

    elif base_inputs:
        inputs = base_inputs
    elif test_case.input_overrides:
        inputs = test_case.input_overrides
    else:
        inputs = celtypes.MapType()

    if test_case.resource_overlay:
        if not current_resource:
            return TestCaseOutcome(
                new_inputs=base_inputs,
                new_resource=current_resource,
                result=TestCaseResult(
                    test_pass=False,
                    expected_outcome=None,
                    outcome=None,
                    message="Can not overlay until the full resource exists. Try setting `currentResource` or reordering test cases.",
                    label=test_case.label,
                ),
                fatal_error=True,
            )

        match evaluate_overlay(
            overlay=test_case.resource_overlay,
            inputs=inputs,
            base=current_resource,
            location=f"{test_case.label}:overlayResource",
        ):
            case result.PermFail() as err:
                return TestCaseOutcome(
                    new_inputs=base_inputs,
                    new_resource=current_resource,
                    result=TestCaseResult(
                        test_pass=False,
                        expected_outcome=None,
                        outcome=err,
                        message=f"Could not run test case due to resource overlay error: ({err})",
                        label=test_case.label,
                    ),
                    fatal_error=not test_case.variant,
                )
            case _ as cel_resource:
                resource = convert_bools(cel_resource)

    elif test_case.current_resource:
        resource = test_case.current_resource
    else:
        resource = current_resource

    api = MockApi(current_resource=resource)

    match function_under_test:
        case ResourceFunction():
            reconcile_result, _ = await reconcile_resource_function(
                api=api,
                location=location,
                function=function_under_test,
                owner=(
                    f"{function_under_test.crud_config.resource_api.namespace}",
                    celtypes.MapType({"uid": "unit-test-123"}),
                ),
                inputs=inputs,
            )
        case ValueFunction():
            if not resource:
                base_resource = celtypes.MapType({})
            else:
                base_resource = celpy.json_to_cel(resource)
                assert isinstance(base_resource, celtypes.MapType)

            reconcile_result = await reconcile_value_function(
                location=location,
                function=function_under_test,
                inputs=inputs,
                value_base=base_resource,
            )

    match test_case.assertion:
        case ExpectOutcome(outcome=expected_outcome):
            outcome_result = _validate_outcome_match(
                expected=expected_outcome, actual=reconcile_result
            )
        case ExpectReturn(value=expected_value):
            outcome_result = _validate_return_match(
                expected=expected_value, actual=reconcile_result
            )
        case ExpectResource(resource=expected_resource):
            outcome_result = _validate_resource_match(
                expected=expected_resource,
                materialized=api.materialized,
                actual_outcome=reconcile_result,
            )
        case ExpectDelete(expected=expected):
            passed = api._delete_called == expected

            if api._delete_called:
                message = "Delete called"
            else:
                message = "Delete not called."

            outcome_result = TestCaseResult(
                test_pass=passed,
                expected_outcome=result.Retry(delay=0, message=""),
                outcome=reconcile_result,
                message=message,
            )

    if test_case.variant:
        return TestCaseOutcome(
            new_inputs=base_inputs,
            new_resource=current_resource,
            result=copy.replace(outcome_result, label=test_case.label),
        )

    return TestCaseOutcome(
        new_inputs=inputs,
        new_resource=api.materialized if api._api_called else resource,
        result=copy.replace(outcome_result, label=test_case.label),
        fatal_error=not outcome_result.test_pass,
    )


def _validate_return_match(
    expected: dict | None, actual: result.UnwrappedOutcome[celtypes.Value] | None
):
    if not result.is_unwrapped_ok(actual):
        return TestCaseResult(
            test_pass=False,
            expected_outcome=result.Ok(expected),
            outcome=actual,
            message=f"Non-Ok return value ({actual})",
        )

    converted_actual = convert_bools(actual)

    match = _validate_match(target=expected, actual=converted_actual)
    return TestCaseResult(
        test_pass=match.match,
        expected_outcome=result.Ok(expected),
        outcome=converted_actual,
        differences=match.differences,
    )


def _validate_resource_match(
    expected: dict | None,
    materialized: dict | None,
    actual_outcome: result.UnwrappedOutcome[celtypes.Value] | None,
):
    if materialized is None and expected is not None:
        if actual_outcome and not result.is_unwrapped_ok(actual_outcome):
            outcome_message = actual_outcome.message
        else:
            outcome_message = "unexpected Ok result"

        return TestCaseResult(
            test_pass=False,
            expected_outcome=result.Ok(expected),
            outcome=actual_outcome,
            message=f"No resource changes attempted: {outcome_message}",
        )

    if not isinstance(actual_outcome, result.Retry):
        return TestCaseResult(
            test_pass=False,
            expected_outcome=result.Ok(expected),
            outcome=actual_outcome,
            message=f"Unexpected result-type ({actual_outcome}), but expected Retry",
        )

    materialized_without_last_applied = _strip_last_applied_annotation(materialized)

    match = _validate_match(target=expected, actual=materialized_without_last_applied)
    return TestCaseResult(
        test_pass=match.match,
        expected_outcome=result.Ok(expected),
        outcome=actual_outcome,
        differences=match.differences,
    )


def _validate_outcome_match(
    expected: result.Outcome | None,
    actual: result.UnwrappedOutcome[celtypes.Value] | None,
):
    test_pass = False
    message = None

    match (expected, actual):
        case result.Retry(message=expected_message, delay=expected_delay), result.Retry(
            message=actual_message, delay=actual_delay
        ):
            if not expected_message:
                test_pass = True
            elif actual_message and (
                expected_message.lower() in actual_message.lower()
            ):
                test_pass = True

            if expected_delay and (expected_delay != actual_delay):
                # test_pass must be True here, so it needs set to False on mismatch
                test_pass = False

            if not test_pass:
                message = (
                    "Retry: "
                    f"Expected(message={expected_message}, delay={expected_delay}) "
                    f"Actual(message={actual_message}, delay={actual_delay})"
                )

        case result.PermFail(message=expected_message), result.PermFail(
            message=actual_message
        ):
            if not expected_message:
                test_pass = True
            elif actual_message and (
                expected_message.lower() in actual_message.lower()
            ):
                test_pass = True
            else:
                message = (
                    "PermFail: "
                    f"Expected(message={expected_message}) "
                    f"Actual(message={actual_message})"
                )

        case result.DepSkip(message=expected_message), result.DepSkip(
            message=actual_message
        ):
            if not expected_message:
                test_pass = True
            elif actual_message and (
                expected_message.lower() in actual_message.lower()
            ):
                test_pass = True
            else:
                message = (
                    "DepSkip: "
                    f"Expected(message={expected_message}) "
                    f"Actual(message={actual_message})"
                )

        case result.Skip(message=expected_message), result.Skip(message=actual_message):
            if not expected_message:
                test_pass = True
            elif actual_message and (
                expected_message.lower() in actual_message.lower()
            ):
                test_pass = True
            else:
                message = (
                    "Skip: "
                    f"Expected(message={expected_message}) "
                    f"Actual(message={actual_message})"
                )

        case _, _:
            if actual and result.is_unwrapped_ok(actual):
                actual = convert_bools(actual)
                actual_str = f"Ok({actual})"
            else:
                actual_str = f"{actual}"

            if expected and result.is_ok(expected):
                if not expected.data:
                    test_pass = True
                else:
                    match = _validate_match(expected.data, actual)
                    test_pass = match.match
                    if not test_pass:
                        message = f"Mismatch at {" -> ".join(match.differences)}"
            elif expected is None:
                test_pass = result.is_unwrapped_ok(actual)
                if not test_pass:
                    message = f"Expected(Ok) Actual({actual_str})"

            else:
                message = f"Expected({expected}) Actual({actual_str})"

    return TestCaseResult(
        test_pass=test_pass,
        expected_outcome=expected,
        outcome=actual,
        message=message,
    )


Difference = dict[str, "str | Difference"] | Sequence["Difference | None"] | str


class ResourceMatch(NamedTuple):
    match: bool
    differences: Difference


def _strip_last_applied_annotation[T](actual: T) -> T:
    if not actual or not isinstance(actual, dict):
        return actual

    if "metadata" not in actual:
        return actual

    if "annotations" not in actual["metadata"]:
        return actual

    annotation_count = len(actual["metadata"]["annotations"])
    if annotation_count == 0:
        return actual

    stripped = copy.deepcopy(actual)
    if annotation_count == 1:
        del stripped["metadata"]["annotations"]
    else:
        del stripped["metadata"]["annotations"][LAST_APPLIED_ANNOTATION]

    return stripped


def _validate_match(target, actual, compare_list_as_set: bool = False) -> ResourceMatch:
    """Compare the specified (`target`) state against the actual (`actual`)
    reosurce state. We compare all target fields and ignore anything extra.
    """
    match (target, actual):
        # Objects need a special comparator
        case dict(), dict():
            return _validate_dict_match(target, actual)
        case dict(), _:
            return ResourceMatch(
                match=False,
                differences=f"expected `object` but found `{type(actual)}`",
            )
        case _, dict():
            return ResourceMatch(
                match=False,
                differences=f"expected `{type(target)}` but found `object`",
            )

        # Arrays need a special comparator
        case (list() | tuple(), list() | tuple()):
            if compare_list_as_set:
                return _validate_set_match(target, actual)

            return _validate_list_match(target, actual)

        case (list() | tuple(), _):
            return ResourceMatch(
                match=False,
                differences=f"expected `array` but found `{type(actual)}`",
            )
        case (_, list() | tuple()):
            return ResourceMatch(
                match=False,
                differences=f"expected `{type(target)}` but found `array`",
            )

        # Bool needs a special comparator, due to Python's int truthiness rules
        case bool(), bool():
            if target == actual:
                return ResourceMatch(match=True, differences=())
            else:
                return ResourceMatch(
                    match=False,
                    differences=f"expected `{target}` but found `{actual}`",
                )
        case bool(), _:
            return ResourceMatch(
                match=False,
                differences=f"expected `bool` but found `{type(actual)}`",
            )
        case _, bool():
            return ResourceMatch(
                match=False,
                differences=f"expected `{type(target)}` but found `bool`",
            )

    # Hopefully anything else is a simple type.
    if target == actual:
        return ResourceMatch(match=True, differences=())

    return ResourceMatch(
        match=False,
        differences=f"expected '`{target}`' but found '`{actual}`'",
    )


KOREO_DIRECTIVE_KEYS: set[str] = {
    "x-koreo-compare-as-set",
    "x-koreo-compare-as-map",
}


def _obj_to_key(obj: dict, fields: Sequence[int | str]) -> str:
    return "$".join(f"{obj.get(field)}".strip() for field in fields)


def _list_to_object(
    obj_list: Sequence[dict], key_fields: Sequence[int | str]
) -> dict[str, dict]:
    return {_obj_to_key(obj, key_fields): obj for obj in obj_list}


def _validate_dict_match(target: dict, actual: dict) -> ResourceMatch:
    differences: dict[str, str | Difference] = {}

    compare_list_as_set_keys = {
        key for key in target.get("x-koreo-compare-as-set", ()) if key
    }

    compare_as_map = {
        key: [field_name for field_name in fields if field_name]
        for key, fields in target.get("x-koreo-compare-as-map", {}).items()
        if key
    }

    target_keys = set(target.keys()) - KOREO_DIRECTIVE_KEYS
    actual_keys = set(actual.keys()) - KOREO_DIRECTIVE_KEYS

    for missing_key in target_keys - actual_keys:
        differences[missing_key] = "missing"

    for unexpected_key in actual_keys - target_keys:
        differences[unexpected_key] = "unexpected"

    for compare_key in target_keys.intersection(actual_keys):
        if compare_key in compare_as_map:
            key_match = _validate_match(
                _list_to_object(target[compare_key], compare_as_map[compare_key]),
                _list_to_object(actual[compare_key], compare_as_map[compare_key]),
            )
        else:
            key_match = _validate_match(
                target[compare_key],
                actual[compare_key],
                compare_list_as_set=(compare_key in compare_list_as_set_keys),
            )

        if not key_match.match:
            differences[compare_key] = key_match.differences

    return ResourceMatch(match=not differences, differences=differences)


def _validate_list_match(target: list | tuple, actual: list | tuple) -> ResourceMatch:
    if len(target) != len(actual):
        return ResourceMatch(
            match=False,
            differences=f"<length mismatch target:{len(target)}, actual:{len(actual)}",
        )

    has_differences = False
    differences = []
    for target_value, actual_value in zip(target, actual):
        item_match = _validate_match(target_value, actual_value)
        if item_match.match:
            differences.append(None)
        else:
            has_differences = True
            differences.append(item_match.differences)

    return ResourceMatch(match=not has_differences, differences=differences)


def _validate_set_match(target: list | tuple, actual: list | tuple) -> ResourceMatch:
    try:
        target_set = set(target)
        actual_set = set(actual)
    except TypeError as err:
        if "dict" in f"{err}":
            return ResourceMatch(
                match=False,
                differences=f"Could not compare array-as-set, try: `x-koreo-compare-as-map`",
            )

        return ResourceMatch(
            match=False, differences=f"Could not compare array-as-set ({err})"
        )
    except Exception as err:
        return ResourceMatch(
            match=False, differences=f"Could not compare array-as-set ({err})"
        )

    missing_values = target_set - actual_set
    unexpected_values = actual_set - target_set

    if not (missing_values or unexpected_values):
        return ResourceMatch(match=True, differences=())

    differences = []

    for missing_value in missing_values:
        differences.append(
            f"<missing '{missing_value}'>",
        )

    for unexpected_value in unexpected_values:
        differences.append(
            f"<unexpectedly found '{unexpected_value}'>",
        )

    return ResourceMatch(match=False, differences=differences)



import random
import string
import unittest

from koreo.workflow import registry


def _name_generator(prefix: str):
    return f"{prefix}-{"".join(random.choices(string.ascii_lowercase, k=15))}"


class TestRegistry(unittest.TestCase):
    def setUp(self):
        registry._reset_registry()

    def test_roundtrip(self):
        custom_crd_name = _name_generator("crd")
        workflow_name_one = _name_generator("workflow")
        workflow_name_two = _name_generator("workflow")

        registry.index_workflow_custom_crd(
            workflow=workflow_name_one, custom_crd=custom_crd_name
        )
        registry.index_workflow_custom_crd(
            workflow=workflow_name_two, custom_crd=custom_crd_name
        )

        workflows = registry.get_custom_crd_workflows(custom_crd=custom_crd_name)
        self.assertIn(workflow_name_one, workflows)
        self.assertIn(workflow_name_two, workflows)

    def test_no_change_usage_reindex(self):
        workflow_name = _name_generator("workflow")

        custom_crd_name = _name_generator("crd")

        registry.index_workflow_custom_crd(
            workflow=workflow_name, custom_crd=custom_crd_name
        )
        registry.index_workflow_custom_crd(
            workflow=workflow_name, custom_crd=custom_crd_name
        )

        self.assertEqual(
            [workflow_name], registry.get_custom_crd_workflows(custom_crd_name)
        )

    def test_change_usage(self):
        workflow_name = _name_generator("workflow")

        custom_crd_name_one = _name_generator("crd")
        custom_crd_name_two = _name_generator("crd")

        registry.index_workflow_custom_crd(
            workflow=workflow_name, custom_crd=custom_crd_name_one
        )
        registry.index_workflow_custom_crd(
            workflow=workflow_name, custom_crd=custom_crd_name_two
        )

        self.assertNotIn(
            workflow_name, registry.get_custom_crd_workflows(custom_crd_name_one)
        )
        self.assertIn(
            workflow_name, registry.get_custom_crd_workflows(custom_crd_name_two)
        )



import logging

logger = logging.getLogger("koreo.resourcetemplate.prepare")

import celpy
from celpy import celtypes

from koreo import schema
from koreo.result import PermFail, UnwrappedOutcome

from . import structure


async def prepare_resource_template(
    cache_key: str, spec: dict
) -> UnwrappedOutcome[tuple[structure.ResourceTemplate, None]]:
    logger.debug(f"Prepare resource template {cache_key}")

    if error := schema.validate(
        resource_type=structure.ResourceTemplate, spec=spec, validation_required=True
    ):
        return PermFail(
            error.message,
            location=f"prepare:ResourceTemplate:{cache_key}",
        )

    template_spec = spec.get("template", {})
    template = celpy.json_to_cel(template_spec)
    if not template_spec:
        return PermFail(
            message=f"Missing `spec.template` for ResourceTemplate '{cache_key}'.",
            location=f"prepare:ResourceTemplate:{cache_key}",
        )

    if not isinstance(template, celtypes.MapType):
        return PermFail(
            message=f"ResourceTemplate '{cache_key}' `spec.template` must be an object.",
            location=f"prepare:ResourceTemplate:{cache_key}",
        )

    if not (template_spec.get("apiVersion") and template_spec.get("kind")):
        return PermFail(
            message=(
                f"ResourceTemplate '{cache_key}' `apiVersion` and `kind` must "
                "be set within `spec.template` ("
                f"apiVersion: '{template_spec.get("apiVersion")}', "
                f"kind: '{template_spec.get("kind")}') "
            ),
            location=f"prepare:ResourceTemplate:{cache_key}",
        )

    context = celpy.json_to_cel(spec.get("context", {}))
    if not isinstance(context, celtypes.MapType):
        return PermFail(
            message=f"ResourceTemplate '{cache_key}' `spec.context` ('{context}') must be an object.",
            location=f"prepare:ResourceTemplate:{cache_key}",
        )

    return (
        structure.ResourceTemplate(
            context=context,
            template=template,
        ),
        None,
    )



from __future__ import annotations
from typing import NamedTuple, Sequence

import celpy

from koreo.result import NonOkOutcome, Outcome

from koreo.resource_function.structure import ResourceFunction
from koreo.value_function.structure import ValueFunction


class ConfigCRDRef(NamedTuple):
    api_group: str
    version: str
    kind: str


class StepConditionSpec(NamedTuple):
    type_: str
    name: str


class LogicSwitch(NamedTuple):
    switch_on: celpy.Runner
    logic_map: dict[str | int, ResourceFunction | ValueFunction | Workflow]
    default_logic: ResourceFunction | ValueFunction | Workflow | None

    dynamic_input_keys: set[str]


class Step(NamedTuple):
    label: str
    logic: ResourceFunction | ValueFunction | Workflow | LogicSwitch

    skip_if: celpy.Runner | None
    for_each: ForEach | None
    inputs: celpy.Runner | None

    condition: StepConditionSpec | None
    state: celpy.Runner | None

    dynamic_input_keys: set[str]


class ForEach(NamedTuple):
    source_iterator: celpy.Runner
    input_key: str
    condition: StepConditionSpec | None


class ErrorStep(NamedTuple):
    label: str
    outcome: NonOkOutcome

    condition: StepConditionSpec | None
    state: None = None


class Workflow(NamedTuple):
    name: str
    crd_ref: ConfigCRDRef | None

    steps_ready: Outcome
    steps: Sequence[Step | ErrorStep]

    dynamic_input_keys: set[str]



from typing import Sequence
from collections import defaultdict
import asyncio
import random
import string
import unittest

from koreo.cache import (
    delete_from_cache,
    get_resource_from_cache,
    prepare_and_cache,
    _reset_cache,
    _REPREPARE_TASKS,
)

from koreo import result
from koreo import registry


def _name_generator():
    return "".join(random.choices(string.ascii_lowercase, k=15))


async def dangerous_prepare(key: str, value_spec: dict):
    raise Exception("Unexpected Prepare!")


async def failed_prepare(key: str, value_spec: dict):
    return result.PermFail("Failed to prepare.")


class ResourceTypeOne:
    prep_count: int
    name: str

    def __init__(self, **values):
        self.prep_count = values.get("prep_count", 0)
        self.name = values.get("name", "----MISSING----")

    def __str__(self):
        return f"<{self.__class__.__name__} prep_count={self.prep_count}>"


_PREP_COUNTER: dict[str, int] = defaultdict(int)


async def prepare_type_one(key: str, value_spec: dict):
    _PREP_COUNTER[key] += 1
    value_spec["prep_count"] = _PREP_COUNTER[key]

    return ResourceTypeOne(**value_spec), None


def build_preparer(resources: Sequence[registry.Resource]):
    async def preparer(key: str, value_spec: dict):
        _PREP_COUNTER[key] += 1
        value_spec["prep_count"] = _PREP_COUNTER[key]

        return ResourceTypeOne(**value_spec), resources

    return preparer


def build_timebomb_preparer(
    detonate_after: int, resources: Sequence[registry.Resource]
):
    async def preparer(key: str, value_spec: dict):
        # This is insanely dirty, never do this for real.
        _PREP_COUNTER[key] += 1
        value_spec["prep_count"] = _PREP_COUNTER[key]

        if _PREP_COUNTER[key] > detonate_after:
            raise Exception(f"Failed after '{detonate_after}' reprepares.")

        return ResourceTypeOne(**value_spec), resources

    return preparer


class TestCache(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        _reset_cache()
        _PREP_COUNTER.clear()

    async def test_get_missing(self):
        from_cache = get_resource_from_cache(
            resource_class=ResourceTypeOne, cache_key=_name_generator()
        )

        self.assertIsNone(from_cache)

    async def test_prepare_bad_metadata(self):
        metadata = {}

        with self.assertRaisesRegex(TypeError, "resource name and version"):
            await prepare_and_cache(
                resource_class=ResourceTypeOne,
                preparer=prepare_type_one,
                metadata=metadata,
                spec={},
            )

    async def test_roundtrip(self):
        resource_name = _name_generator()
        resource_version = f"v{random.randint(10, 1000000)}"

        metadata = {"name": resource_name, "resourceVersion": resource_version}
        spec = {"name": resource_name}
        prepared = await prepare_and_cache(
            resource_class=ResourceTypeOne,
            preparer=prepare_type_one,
            metadata=metadata,
            spec=spec,
        )

        self.assertEqual(prepared.prep_count, _PREP_COUNTER[resource_name])
        self.assertEqual(prepared.name, resource_name)

        from_cache = get_resource_from_cache(
            resource_class=ResourceTypeOne, cache_key=resource_name
        )

        self.assertEqual(from_cache.prep_count, prepared.prep_count)
        self.assertEqual(from_cache.name, prepared.name)

    async def test_changes_reprepared(self):
        resource_name = _name_generator()
        first_resource_version = f"v{random.randint(1000, 100000)}"

        first_metadata = {
            "name": resource_name,
            "resourceVersion": first_resource_version,
        }

        spec = {"name": resource_name}
        first_prepared = await prepare_and_cache(
            resource_class=ResourceTypeOne,
            preparer=prepare_type_one,
            metadata=first_metadata,
            spec=spec,
        )

        self.assertEqual(first_prepared.prep_count, _PREP_COUNTER[resource_name])
        self.assertEqual(first_prepared.name, resource_name)

        second_resource_version = f"v{random.randint(100, 999)}"

        second_metadata = {
            "name": resource_name,
            "resourceVersion": second_resource_version,
        }
        second_prepared = await prepare_and_cache(
            resource_class=ResourceTypeOne,
            preparer=prepare_type_one,
            metadata=second_metadata,
            spec=spec,
        )

        self.assertEqual(second_prepared.prep_count, _PREP_COUNTER[resource_name])
        self.assertEqual(first_prepared.prep_count + 1, second_prepared.prep_count)
        self.assertEqual(first_prepared.name, second_prepared.name)

    async def test_prepare_fail(self):
        resource_name = _name_generator()
        resource_version = f"v{random.randint(10, 1000000)}"

        metadata = {"name": resource_name, "resourceVersion": resource_version}

        spec = {"name": resource_name}
        first_prepared = await prepare_and_cache(
            resource_class=ResourceTypeOne,
            preparer=failed_prepare,
            metadata=metadata,
            spec=spec,
        )

        self.assertIsInstance(first_prepared, result.PermFail)

    async def test_duplicate_not_reprepared(self):
        resource_name = _name_generator()
        resource_version = f"v{random.randint(10, 1000000)}"

        metadata = {"name": resource_name, "resourceVersion": resource_version}

        spec = {"name": resource_name}
        first_prepared = await prepare_and_cache(
            resource_class=ResourceTypeOne,
            preparer=prepare_type_one,
            metadata=metadata,
            spec=spec,
        )

        self.assertEqual(first_prepared.prep_count, 1)
        self.assertEqual(first_prepared.name, resource_name)

        second_prepared = await prepare_and_cache(
            resource_class=ResourceTypeOne,
            preparer=dangerous_prepare,
            metadata=metadata,
            spec=spec,
        )

        self.assertEqual(_PREP_COUNTER[resource_name], 1)
        self.assertEqual(first_prepared.prep_count, second_prepared.prep_count)
        self.assertEqual(first_prepared.name, second_prepared.name)


class TestCacheRegistry(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        _reset_cache()
        _PREP_COUNTER.clear()

    async def test_repreparer_skip(self):
        # Resource names.
        resource_name = f"dependent-{_name_generator()}"
        watched_resource_name = f"watched-{_name_generator()}"

        # We want a Lifo queue for the test.
        check_queue = asyncio.LifoQueue()
        reg_resource = registry.Resource(
            resource_type=ResourceTypeOne, name=resource_name
        )
        registry.register(registerer=reg_resource, queue=check_queue)

        # Build a resource that watches.
        resource_version = f"v{random.randint(10, 1000000)}"
        metadata = {"name": resource_name, "resourceVersion": resource_version}
        spec = {"name": resource_name}
        await prepare_and_cache(
            resource_class=ResourceTypeOne,
            preparer=build_preparer(
                (
                    registry.Resource(
                        resource_type=ResourceTypeOne, name=watched_resource_name
                    ),
                )
            ),
            metadata=metadata,
            spec=spec,
        )

        # Build a resource that will be watched.
        watched_spec = {"name": watched_resource_name}

        for _ in range(5):
            watched_resource_version = f"v{random.randint(1000, 100000)}"
            watched_metadata = {
                "name": watched_resource_name,
                "resourceVersion": watched_resource_version,
            }
            await prepare_and_cache(
                resource_class=ResourceTypeOne,
                preparer=prepare_type_one,
                metadata=watched_metadata,
                spec=watched_spec,
            )

        await check_queue.join()
        registry.kill_resource(reg_resource)
        await check_queue.join()

        self.assertTrue(check_queue.empty())
        self.assertGreater(_PREP_COUNTER[resource_name], 1)
        self.assertLessEqual(_PREP_COUNTER[resource_name], 5)
        self.assertEqual(_PREP_COUNTER[watched_resource_name], 5)

    async def test_repreparer_failure(self):
        # Resource names.
        resource_name = f"dependent-{_name_generator()}"
        watched_resource_name = f"watched-{_name_generator()}"

        # We want a non-Lifo Queue for the test.
        check_queue = asyncio.Queue()
        reg_resource = registry.Resource(
            resource_type=ResourceTypeOne, name=resource_name
        )
        registry.register(registerer=reg_resource, queue=check_queue)

        # Build a resource that watches.
        resource_version = f"v{random.randint(10, 1000000)}"
        metadata = {"name": resource_name, "resourceVersion": resource_version}
        spec = {"name": resource_name}
        await prepare_and_cache(
            resource_class=ResourceTypeOne,
            preparer=build_timebomb_preparer(
                detonate_after=3,
                resources=(
                    registry.Resource(
                        resource_type=ResourceTypeOne, name=watched_resource_name
                    ),
                ),
            ),
            metadata=metadata,
            spec=spec,
        )

        # Build a resource that will be watched.
        watched_spec = {"name": watched_resource_name}

        for _ in range(5):
            watched_resource_version = f"v{random.randint(1000, 100000)}"
            watched_metadata = {
                "name": watched_resource_name,
                "resourceVersion": watched_resource_version,
            }
            await prepare_and_cache(
                resource_class=ResourceTypeOne,
                preparer=prepare_type_one,
                metadata=watched_metadata,
                spec=watched_spec,
            )

        await check_queue.join()

        self.assertTrue(check_queue.empty())
        self.assertGreater(_PREP_COUNTER[resource_name], 1)
        self.assertLessEqual(_PREP_COUNTER[resource_name], 5)
        self.assertEqual(_PREP_COUNTER[watched_resource_name], 5)

    async def test_delete_cleans_up(self):
        # Resource names.
        resource_name = f"dependent-{_name_generator()}"
        delete_resource_name = f"delete-{_name_generator()}"
        watched_resource_name = f"watched-{_name_generator()}"

        # Build a resource that watches.
        reg_resource = registry.Resource(
            resource_type=ResourceTypeOne, name=resource_name
        )
        reg_queue = registry.register(registerer=reg_resource)

        await prepare_and_cache(
            resource_class=ResourceTypeOne,
            preparer=build_preparer(
                resources=(
                    registry.Resource(
                        resource_type=ResourceTypeOne, name=delete_resource_name
                    ),
                ),
            ),
            metadata={
                "name": resource_name,
                "resourceVersion": f"v{random.randint(1000, 100000)}",
            },
            spec={"name": resource_name},
        )

        # Build a resource to delete that watches something else.
        del_resource = registry.Resource(
            resource_type=ResourceTypeOne, name=delete_resource_name
        )
        del_queue = registry.register(registerer=del_resource)
        await prepare_and_cache(
            resource_class=ResourceTypeOne,
            preparer=build_preparer(
                resources=(
                    registry.Resource(
                        resource_type=ResourceTypeOne, name=watched_resource_name
                    ),
                ),
            ),
            metadata={
                "name": delete_resource_name,
                "resourceVersion": f"v{random.randint(1000, 100000)}",
            },
            spec={"name": delete_resource_name},
        )

        # Build a resource that will be watched, and update it a few times.
        watched_spec = {"name": watched_resource_name}
        for _ in range(5):
            watched_resource_version = f"v{random.randint(1000, 100000)}"
            watched_metadata = {
                "name": watched_resource_name,
                "resourceVersion": watched_resource_version,
            }
            await prepare_and_cache(
                resource_class=ResourceTypeOne,
                preparer=prepare_type_one,
                metadata=watched_metadata,
                spec=watched_spec,
            )

        await del_queue.join()
        await reg_queue.join()

        # I'm not a huge fan of reaching in so deeply, but in this case we
        # want to make sure we cleared the task.

        self.assertEqual(len(_REPREPARE_TASKS), 2)

        to_be_deleted = get_resource_from_cache(
            resource_class=ResourceTypeOne, cache_key=delete_resource_name
        )
        self.assertIsNotNone(to_be_deleted)

        await delete_from_cache(
            resource_class=ResourceTypeOne, cache_key=delete_resource_name
        )

        await del_queue.join()
        await reg_queue.join()

        deleted = get_resource_from_cache(
            resource_class=ResourceTypeOne, cache_key=delete_resource_name
        )

        self.assertEqual(len(_REPREPARE_TASKS), 1)
        self.assertIsNone(deleted)



from typing import Any
import base64

from celpy import celtypes

CEL_PREFIX = "="

ConvertedType = (
    list["ConvertedType"]
    | dict["ConvertedType", "ConvertedType"]
    | str
    | int
    | float
    | bool
    | None
)


def convert_bools(
    cel_object: celtypes.Value,
) -> ConvertedType:
    """Recursive walk through the CEL object, replacing celtypes with native
    types. This lets the :py:mod:`json` module correctly represent the obects
    and allows Python code to treat these as normal objects.
    """
    match cel_object:
        case celtypes.BoolType():
            return True if cel_object else False

        case celtypes.StringType() | celtypes.TimestampType() | celtypes.DurationType():
            return str(cel_object)

        case celtypes.IntType() | celtypes.UintType():
            return int(cel_object)

        case celtypes.DoubleType():
            return float(cel_object)

        case celtypes.BytesType():
            return base64.b64encode(cel_object).decode("ASCII")

        case celtypes.ListType() | list() | tuple():
            return [convert_bools(item) for item in cel_object]

        case celtypes.MapType() | dict():
            return {
                convert_bools(key): convert_bools(value)
                for key, value in cel_object.items()
            }

        case celtypes.NullType():
            return None

        case _:
            return cel_object


def encode_cel(value: Any) -> str:
    if isinstance(value, dict):
        return f"{{{ ",".join(
            f'"{f"{key}".replace('"', '\"')}":{encode_cel(value)}'
            for key, value in value.items()
        ) }}}"

    if isinstance(value, list):
        return f"[{ ",".join(encode_cel(item) for item in value) }]"

    if isinstance(value, bool):
        return "true" if value else "false"

    if value is None:
        return "null"

    if _encode_plain(value):
        return f"{value}"

    if not value:
        return '""'

    if not value.startswith(CEL_PREFIX):
        if "\n" in value:
            return f'r"""{ value.replace('"', '\"') }"""'  # fmt: skip

        if '"' in value:
            return f'"""{ value.replace('"', r'\"') }"""'  # fmt: skip

        return f'"{value}"'  # fmt: skip

    return value.lstrip(CEL_PREFIX)


def _encode_plain(maybe_number) -> bool:
    if not isinstance(maybe_number, str):
        return True

    try:
        int(maybe_number)
        return True
    except ValueError:
        pass

    try:
        float(maybe_number)
        return True
    except ValueError:
        pass

    return False



from collections import defaultdict
from typing import NamedTuple, Sequence
import asyncio
import time
import logging

logger = logging.getLogger(name="koreo.registry")


class Resource[T](NamedTuple):
    resource_type: type[T]
    name: str
    namespace: str | None = None


class Kill: ...


class ResourceEvent[T](NamedTuple):
    resource: Resource[T]
    event_time: float


type RegistryQueue = asyncio.Queue[ResourceEvent | Kill]


def register[
    T
](registerer: Resource[T], queue: RegistryQueue | None = None,) -> RegistryQueue:
    registerer_key = _resource_key(registerer)

    if registerer_key in _SUBSCRIPTION_QUEUES:
        return _SUBSCRIPTION_QUEUES[registerer_key]

    if not queue:
        queue = asyncio.LifoQueue[ResourceEvent | Kill]()

    _SUBSCRIPTION_QUEUES[registerer_key] = queue

    event_time = time.monotonic()
    notify_subscribers(notifier=registerer, event_time=event_time)

    logger.info(f"Registering {registerer}")

    return queue


class SubscriptionCycle(Exception): ...


def subscribe(subscriber: Resource, resource: Resource):
    subscriber_key = _resource_key(subscriber)
    resource_key = _resource_key(resource)

    _check_for_cycles(subscriber_key, (resource_key,))

    _RESOURCE_SUBSCRIBERS[resource_key].add(subscriber_key)
    _SUBSCRIBER_RESOURCES[subscriber_key].add(resource_key)

    logger.info(f"{subscriber} subscribing to {resource}")


def subscribe_only_to(subscriber: Resource, resources: Sequence[Resource]):
    subscriber_key = _resource_key(subscriber)

    new = set(_resource_key(subscribe_to) for subscribe_to in resources)
    _check_for_cycles(subscriber_key, list(new))

    current = _SUBSCRIBER_RESOURCES[subscriber_key]

    for resource_key in new - current:
        _RESOURCE_SUBSCRIBERS[resource_key].add(subscriber_key)

    for resource_key in current - new:
        _RESOURCE_SUBSCRIBERS[resource_key].remove(subscriber_key)

    _SUBSCRIBER_RESOURCES[subscriber_key] = new

    logger.info(f"{subscriber} subscribing to {resources}")


def unsubscribe(unsubscriber: Resource, resource: Resource):
    unsubscriber_key = _resource_key(unsubscriber)
    resource_key = _resource_key(resource)

    _RESOURCE_SUBSCRIBERS[resource_key].remove(unsubscriber_key)
    _SUBSCRIBER_RESOURCES[unsubscriber_key].remove(resource_key)


def notify_subscribers(notifier: Resource, event_time: float):
    resource_key = _resource_key(notifier)
    subscribers = _RESOURCE_SUBSCRIBERS[resource_key]
    if not subscribers:
        logger.info(f"{notifier} has no subscribers")
        return

    active_subscribers = [
        _SUBSCRIPTION_QUEUES[subscriber]
        for subscriber in subscribers
        if subscriber in _SUBSCRIPTION_QUEUES
    ]

    if not active_subscribers:
        logger.info(f"{notifier} has no active subscribers")
        return

    logger.info(f"{notifier}:{event_time} notifying to {subscribers}")

    for subscriber in active_subscribers:
        try:
            subscriber.put_nowait(
                ResourceEvent(resource=notifier, event_time=event_time)
            )
        except asyncio.QueueFull:
            pass
            # TODO: I think there is a way to monitor for stalled subscribers
            # then notify a house-keeper process to deal with it.

            # health_check_task = asyncio.create_task()
            # _CHECK_SUBSCRIBER_HEALTH.add(health_check_task)
            # health_check_task.add_done_callback(_CHECK_SUBSCRIBER_HEALTH.discard)


def get_subscribers(resource: Resource):
    resource_key = _resource_key(resource)

    return _RESOURCE_SUBSCRIBERS[resource_key]


def kill_resource(resource: Resource) -> RegistryQueue | None:
    resource_key = _resource_key(resource)
    if resource_key not in _SUBSCRIPTION_QUEUES:
        return None

    _kill_resource(resource_key)


def deregister(deregisterer: Resource, deregistered_at: float):
    deregisterer_key = _resource_key(deregisterer)

    # This resource is no longer following any resources.
    subscribe_only_to(subscriber=deregisterer, resources=[])

    # Remove this resource's subscription queue
    if deregisterer_key in _SUBSCRIPTION_QUEUES:
        queue = _kill_resource(resource_key=deregisterer_key)
        assert queue  # Just for the type-checker

        del _SUBSCRIPTION_QUEUES[deregisterer_key]

        # This is to prevent blocking anything waiting for this resource to do
        # something.
        while not queue.empty():
            try:
                queue.get_nowait()
                queue.task_done()
            except asyncio.QueueEmpty:
                break

    # Inform subscribers of a change
    notify_subscribers(notifier=deregisterer, event_time=deregistered_at)


class _ResourceKey(NamedTuple):
    resource_type: str
    name: str
    namespace: str | None = None


def _resource_key(resource: Resource) -> _ResourceKey:
    return _ResourceKey(
        resource_type=resource.resource_type.__qualname__,
        name=resource.name,
        namespace=resource.namespace,
    )


def _kill_resource(resource_key: _ResourceKey) -> RegistryQueue | None:
    queue = _SUBSCRIPTION_QUEUES[resource_key]
    try:
        queue.put_nowait(Kill())

    except asyncio.QueueShutDown:
        return queue

    except asyncio.QueueFull:
        pass

    queue.shutdown()

    return queue


def _check_for_cycles(
    subscriber_key: _ResourceKey, resource_keys: Sequence[_ResourceKey]
):
    # Simple, inefficient cycle detection. This is a simple brute-force check,
    # which hopefully given the problem space is sufficient.
    checked: set[_ResourceKey] = set()
    to_check: set[_ResourceKey] = set(resource_keys)
    while True:
        if not to_check:
            break

        if subscriber_key in to_check:
            raise SubscriptionCycle(
                f"Detected subscription cycle due to {subscriber_key}"
            )

        checked.update(to_check)

        next_check_set = set[_ResourceKey]()
        for check_resource_key in to_check:
            if check_resource_key not in _SUBSCRIBER_RESOURCES:
                continue

            check_resource_subscriptions = _SUBSCRIBER_RESOURCES[check_resource_key]
            next_check_set.update(check_resource_subscriptions)
        to_check = next_check_set


_RESOURCE_SUBSCRIBERS: defaultdict[_ResourceKey, set[_ResourceKey]] = defaultdict(
    set[_ResourceKey]
)
_SUBSCRIBER_RESOURCES: defaultdict[_ResourceKey, set[_ResourceKey]] = defaultdict(
    set[_ResourceKey]
)
_SUBSCRIPTION_QUEUES: dict[_ResourceKey, RegistryQueue] = {}


def _reset_registries():
    _RESOURCE_SUBSCRIBERS.clear()
    _SUBSCRIBER_RESOURCES.clear()

    for queue in _SUBSCRIPTION_QUEUES.values():
        try:
            queue.put_nowait(Kill())
        except (asyncio.QueueFull, asyncio.QueueShutDown):
            pass

        try:
            while not queue.empty():
                queue.get_nowait()
                queue.task_done()
        except asyncio.QueueEmpty:
            pass

    _SUBSCRIPTION_QUEUES.clear()



import unittest

import celpy
from celpy import celtypes

from koreo.result import Ok, is_unwrapped_ok

from koreo.cel.prepare import prepare_overlay_expression

from koreo.value_function import structure as function_structure

from koreo.workflow import reconcile
from koreo.workflow import structure as workflow_structure


class TestReconcileWorkflow(unittest.IsolatedAsyncioTestCase):
    async def test_reconcile(self):
        cel_env = celpy.Environment()
        source_return_value = prepare_overlay_expression(
            cel_env=cel_env,
            spec={"resources": [{"bool": True}, {"bool": False}]},
            location="unittest",
        )

        step_state = cel_env.program(cel_env.compile("{'input_source': value}"))

        workflow = workflow_structure.Workflow(
            name="unit-test",
            crd_ref=workflow_structure.ConfigCRDRef(
                api_group="tests.koreo.dev", version="v1", kind="TestCase"
            ),
            steps_ready=Ok(None),
            steps=[
                workflow_structure.Step(
                    label="input_source",
                    skip_if=None,
                    for_each=None,
                    inputs=None,
                    dynamic_input_keys=[],
                    logic=function_structure.ValueFunction(
                        preconditions=None,
                        local_values=None,
                        return_value=source_return_value,
                        dynamic_input_keys=set(),
                    ),
                    condition=None,
                    state=step_state,
                )
            ],
            dynamic_input_keys=set(),
        )

        workflow_result = await reconcile.reconcile_workflow(
            api=None,
            workflow_key="test-case",
            owner=("unit-tests", celtypes.MapType({"uid": "sam-123"})),
            trigger=celtypes.MapType({}),
            workflow=workflow,
        )

        self.maxDiff = None
        self.assertDictEqual(
            {"input_source": {"resources": [{"bool": True}, {"bool": False}]}},
            workflow_result.state,
        )

        # TODO: Check Condition

    async def test_reconcile_nested(self):
        cel_env = celpy.Environment()

        sub_step_return_value = prepare_overlay_expression(
            cel_env=cel_env, spec={"sub_one": True}, location="unittest:sub_step.return"
        )

        assert is_unwrapped_ok(sub_step_return_value), f"{sub_step_return_value}"

        sub_step_two_return_value = prepare_overlay_expression(
            cel_env=cel_env,
            spec={"value": "17171"},
            location="unittest:sub_step_two.return",
        )

        assert is_unwrapped_ok(
            sub_step_two_return_value
        ), f"{sub_step_two_return_value}"

        sub_workflow = workflow_structure.Workflow(
            name="unit-test",
            crd_ref=workflow_structure.ConfigCRDRef(
                api_group="tests.koreo.dev", version="v1", kind="TestCase"
            ),
            steps_ready=Ok(None),
            steps=[
                workflow_structure.Step(
                    label="sub_step",
                    skip_if=None,
                    for_each=None,
                    inputs=None,
                    dynamic_input_keys=[],
                    logic=function_structure.ValueFunction(
                        preconditions=None,
                        local_values=None,
                        return_value=sub_step_return_value,
                        dynamic_input_keys=set(),
                    ),
                    condition=None,
                    state=cel_env.program(cel_env.compile("{'sub_step': value}")),
                ),
                workflow_structure.Step(
                    label="sub_step_two",
                    skip_if=None,
                    for_each=None,
                    inputs=None,
                    dynamic_input_keys=[],
                    logic=function_structure.ValueFunction(
                        preconditions=None,
                        local_values=None,
                        return_value=sub_step_two_return_value,
                        dynamic_input_keys=set(),
                    ),
                    condition=None,
                    state=cel_env.program(cel_env.compile("{'sub_step_two': value}")),
                ),
            ],
            dynamic_input_keys=set(),
        )

        workflow = workflow_structure.Workflow(
            name="unit-test",
            crd_ref=workflow_structure.ConfigCRDRef(
                api_group="tests.koreo.dev", version="v1", kind="TestCase"
            ),
            steps_ready=Ok(None),
            steps=[
                workflow_structure.Step(
                    label="sub_workflow",
                    skip_if=None,
                    for_each=None,
                    inputs=None,
                    dynamic_input_keys=[],
                    logic=sub_workflow,
                    condition=None,
                    state=cel_env.program(cel_env.compile("{'sub_workflow': value}")),
                )
            ],
            dynamic_input_keys=set(),
        )

        workflow_result = await reconcile.reconcile_workflow(
            api=None,
            workflow_key="test-case",
            owner=("unit-tests", celtypes.MapType({"uid": "sam-123"})),
            trigger=celtypes.MapType({}),
            workflow=workflow,
        )
        print(workflow_result)

        self.maxDiff = None
        self.assertDictEqual(
            {
                "sub_workflow": {
                    "sub_step": {"sub_one": True},
                    "sub_step_two": {"value": 17171},
                }
            },
            workflow_result.state,
        )

        # TODO: Check Condition

    async def test_partial_state(self):
        cel_env = celpy.Environment()

        workflow = workflow_structure.Workflow(
            name="unit-test",
            crd_ref=workflow_structure.ConfigCRDRef(
                api_group="tests.koreo.dev", version="v1", kind="TestCase"
            ),
            steps_ready=Ok(None),
            steps=[
                workflow_structure.Step(
                    label="ok_step_one",
                    skip_if=None,
                    for_each=None,
                    inputs=None,
                    dynamic_input_keys=[],
                    logic=function_structure.ValueFunction(
                        preconditions=None,
                        local_values=None,
                        return_value=prepare_overlay_expression(
                            cel_env=cel_env, spec={"i_am_ok": True}, location="unittest"
                        ),
                        dynamic_input_keys=set(),
                    ),
                    condition=None,
                    state=cel_env.program(cel_env.compile("{'first': value}")),
                ),
                workflow_structure.Step(
                    label="fail_step",
                    skip_if=None,
                    for_each=None,
                    inputs=None,
                    dynamic_input_keys=[],
                    logic=function_structure.ValueFunction(
                        preconditions=None,
                        local_values=None,
                        return_value=cel_env.program(cel_env.compile("1 / 0")),
                        dynamic_input_keys=set(),
                    ),
                    condition=None,
                    state=cel_env.program(cel_env.compile("{'failed_step': value}")),
                ),
                workflow_structure.Step(
                    label="ok_step_two",
                    skip_if=None,
                    for_each=None,
                    inputs=None,
                    dynamic_input_keys=[],
                    logic=function_structure.ValueFunction(
                        preconditions=None,
                        local_values=None,
                        return_value=prepare_overlay_expression(
                            cel_env=cel_env, spec={"sub_one": True}, location="unittest"
                        ),
                        dynamic_input_keys=set(),
                    ),
                    condition=None,
                    state=cel_env.program(
                        cel_env.compile("{'number_two': 2, 'two_value': value}")
                    ),
                ),
            ],
            dynamic_input_keys=set(),
        )

        workflow_result = await reconcile.reconcile_workflow(
            api=None,
            workflow_key="test-case",
            owner=("unit-tests", celtypes.MapType({"uid": "sam-123"})),
            trigger=celtypes.MapType({}),
            workflow=workflow,
        )

        self.maxDiff = None
        self.assertDictEqual(
            {
                "first": {"i_am_ok": True},
                "number_two": 2,
                "two_value": {"sub_one": True},
            },
            workflow_result.state,
        )

        # TODO: Check Condition

    async def test_ref_switch(self):
        cel_env = celpy.Environment()
        source_return_value = prepare_overlay_expression(
            cel_env=cel_env,
            spec={"value": "static-value"},
            location="unittest",
        )

        step_state = cel_env.program(cel_env.compile("{'single_switch': value}"))

        workflow = workflow_structure.Workflow(
            name="unit-test",
            crd_ref=workflow_structure.ConfigCRDRef(
                api_group="tests.koreo.dev", version="v1", kind="TestCase"
            ),
            steps_ready=Ok(None),
            steps=[
                workflow_structure.Step(
                    label="single_switch",
                    skip_if=None,
                    for_each=None,
                    inputs=None,
                    dynamic_input_keys=[],
                    logic=workflow_structure.LogicSwitch(
                        switch_on=cel_env.program(cel_env.compile("'alpha'")),
                        logic_map={
                            "alpha": function_structure.ValueFunction(
                                preconditions=None,
                                local_values=None,
                                return_value=source_return_value,
                                dynamic_input_keys=set(),
                            ),
                        },
                        default_logic=None,
                        dynamic_input_keys=set(),
                    ),
                    condition=None,
                    state=step_state,
                )
            ],
            dynamic_input_keys=set(),
        )

        workflow_result = await reconcile.reconcile_workflow(
            api=None,
            workflow_key="test-case",
            owner=("unit-tests", celtypes.MapType({"uid": "sam-123"})),
            trigger=celtypes.MapType({}),
            workflow=workflow,
        )

        self.maxDiff = None
        print(workflow_result.result)
        self.assertDictEqual(
            {"single_switch": {"value": "static-value"}},
            workflow_result.state,
        )

        # TODO: Check Condition


class TestConditionHelper(unittest.TestCase):
    def test_none_outcome(self):
        condition = reconcile._condition_helper(
            condition_type="UnitTest",
            thing_name="unit test",
            outcome=None,
            workflow_key="unit-test-flow",
        )

        self.assertEqual("Pending", condition.get("reason"))

    def test_ok_outcome_that_is_none(self):
        condition = reconcile._condition_helper(
            condition_type="UnitTest",
            thing_name="unit test",
            outcome=Ok(None),
            workflow_key="unit-test-flow",
        )

        self.assertEqual("Ready", condition.get("reason"))

    def test_falsy_outcomes(self):
        falsy_stuff = [[], (), {}, 0, False, ""]

        for falsy_thing in falsy_stuff:
            condition = reconcile._condition_helper(
                condition_type="UnitTest",
                thing_name="unit test",
                outcome=falsy_thing,
                workflow_key="unit-test-flow",
            )

            self.assertEqual(
                "Ready", condition.get("reason"), f"Falsy thing '{falsy_thing}'"
            )



import asyncio
import time
import unittest

from koreo import registry


class ResourceA: ...


class ResourceB: ...


class TestRegistry(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        registry._reset_registries()

    async def test_register_your_own_queue(self):
        resource_a = registry.Resource(resource_type=ResourceA, name="resource-1")
        the_queue = asyncio.LifoQueue()

        a_notifications = registry.register(resource_a, queue=the_queue)

        self.assertIs(a_notifications, the_queue)

    async def test_double_register(self):
        resource_a = registry.Resource(resource_type=ResourceA, name="resource-1")

        a_queue = registry.register(resource_a)

        for _ in range(5):
            self.assertIs(a_queue, registry.register(resource_a))

    async def test_basic_notifications(self):
        resource_a = registry.Resource(resource_type=ResourceA, name="resource-1")
        resource_b = registry.Resource(resource_type=ResourceB, name="resource-2")

        a_notifications = registry.register(resource_a)
        b_notifications = registry.register(resource_b)

        registry.subscribe(resource_a, resource_b)

        event_time = time.monotonic()
        registry.notify_subscribers(resource_b, event_time)

        notification, notification_time = await a_notifications.get()
        self.assertEqual(notification, resource_b)
        self.assertEqual(notification_time, event_time)
        a_notifications.task_done()

        self.assertTrue(a_notifications.empty())
        self.assertTrue(b_notifications.empty())

    async def test_unsubscribe(self):
        resource_a = registry.Resource(resource_type=ResourceA, name="resource-1")
        resource_b = registry.Resource(resource_type=ResourceB, name="resource-2")

        a_notifications = registry.register(resource_a)
        b_notifications = registry.register(resource_b)

        registry.subscribe(resource_a, resource_b)
        registry.unsubscribe(resource_a, resource_b)

        registry.notify_subscribers(resource_b, time.monotonic())

        self.assertTrue(a_notifications.empty())
        self.assertTrue(b_notifications.empty())

    async def test_deregister_subscriber(self):
        resource_a = registry.Resource(resource_type=ResourceA, name="resource-1")
        resource_b = registry.Resource(resource_type=ResourceB, name="resource-2")

        a_notifications = registry.register(resource_a)
        b_notifications = registry.register(resource_b)

        registry.subscribe(resource_a, resource_b)
        for _ in range(10):
            registry.notify_subscribers(resource_b, time.monotonic())

        # Ensure there are pending notifications for A (non for B).
        self.assertFalse(a_notifications.empty())
        self.assertTrue(b_notifications.empty())

        registry.deregister(resource_a, time.monotonic())

        for _ in range(10):
            registry.notify_subscribers(resource_b, time.monotonic())

        # Ensure pending notifications for A are released.
        self.assertTrue(a_notifications.empty())
        self.assertTrue(b_notifications.empty())

        # Ensure re-registering returns a new Queue
        a_notifications_new = registry.register(resource_a)
        self.assertIsNot(a_notifications, a_notifications_new)

    async def test_deregister_resource(self):
        resource_a = registry.Resource(resource_type=ResourceA, name="resource-1")
        resource_b = registry.Resource(resource_type=ResourceB, name="resource-2")

        a_notifications = registry.register(resource_a)
        b_notifications = registry.register(resource_b)

        registry.subscribe(resource_a, resource_b)

        self.assertTrue(a_notifications.empty())
        self.assertTrue(b_notifications.empty())

        registry.deregister(resource_b, time.monotonic())

        notification, _ = await a_notifications.get()
        self.assertEqual(notification, resource_b)
        a_notifications.task_done()

        registry.notify_subscribers(resource_b, time.monotonic())

        # Notifications for the resource should still be sent.
        self.assertFalse(a_notifications.empty())
        self.assertTrue(b_notifications.empty())

    async def test_deregister_unregistered(self):
        resource_a = registry.Resource(resource_type=ResourceA, name="resource-1")
        resource_b = registry.Resource(resource_type=ResourceB, name="resource-2")

        b_notifications = registry.register(resource_b)

        registry.subscribe(resource_a, resource_b)
        for _ in range(10):
            registry.notify_subscribers(resource_b, time.monotonic())

        # Ensure there are pending notifications for A (non for B).
        self.assertTrue(b_notifications.empty())

        registry.deregister(resource_a, time.monotonic())

        for _ in range(10):
            registry.notify_subscribers(resource_b, time.monotonic())

        # Ensure pending notifications for A are released.
        self.assertTrue(b_notifications.empty())

    async def test_changing_subscriptions(self):
        resource_a = registry.Resource(resource_type=ResourceA, name="resource-1")
        resource_b = registry.Resource(resource_type=ResourceB, name="resource-2")
        resource_c = registry.Resource(
            resource_type=ResourceB, name="resource-3", namespace="old"
        )
        resource_d = registry.Resource(
            resource_type=ResourceB, name="resource-4", namespace="new"
        )
        resource_e = registry.Resource(
            resource_type=ResourceB, name="resource-5", namespace="new"
        )

        a_notifications = registry.register(resource_a)
        registry.register(resource_b)
        registry.register(resource_c)
        registry.register(resource_d)
        registry.register(resource_e)

        all_resources = set(
            [resource_a, resource_b, resource_c, resource_d, resource_e]
        )
        first_set = set([resource_b, resource_c])
        second_set = set([resource_b, resource_d, resource_e])

        for resource in first_set:
            registry.subscribe(resource_a, resource)

        for resource in all_resources:
            registry.notify_subscribers(resource, time.monotonic())

        while not a_notifications.empty():
            notifier, _ = await a_notifications.get()
            self.assertIn(notifier, first_set)
            a_notifications.task_done()

        self.assertTrue(a_notifications.empty())

        registry.subscribe_only_to(resource_a, second_set)

        for resource in all_resources:
            registry.notify_subscribers(resource, time.monotonic())

        while not a_notifications.empty():
            notifier, _ = await a_notifications.get()
            self.assertIn(notifier, second_set)
            a_notifications.task_done()

        self.assertTrue(a_notifications.empty())

    async def test_cycle_detection_self(self):
        resource = registry.Resource(resource_type=ResourceA, name="resource-1")
        with self.assertRaises(registry.SubscriptionCycle):
            registry.subscribe(resource, resource)

    async def test_cycle_detection(self):
        all_resources = [
            registry.Resource(resource_type=ResourceA, name=f"resource-{idx}")
            for idx in range(5)
        ]

        for idx, resource in enumerate(all_resources[1:]):
            registry.subscribe(all_resources[idx - 1], resource)

        with self.assertRaises(registry.SubscriptionCycle):
            registry.subscribe(all_resources[-1], all_resources[0])

    async def test_cycle_detection_sub_only_to(self):
        all_resources = [
            registry.Resource(resource_type=ResourceA, name=f"resource-{idx}")
            for idx in range(5)
        ]

        for idx, resource in enumerate(all_resources[:-1]):
            registry.subscribe_only_to(resource, all_resources[idx + 1 :])

        with self.assertRaises(registry.SubscriptionCycle):
            registry.subscribe(all_resources[-1], all_resources[0])
